import logging
from pathlib import Path
from typing import List, Optional

from seml.settings import SETTINGS
from seml.utils import s_if
from seml.utils.errors import MongoDBError
from seml.utils.ssh_forward import get_forwarded_mongo_client

States = SETTINGS.STATES


def get_collection(collection_name, mongodb_config=None, suffix=None):
    if mongodb_config is None:
        mongodb_config = get_mongodb_config()
    db = get_database(**mongodb_config)
    if suffix is not None and not collection_name.endswith(suffix):
        collection_name = f'{collection_name}{suffix}'

    return db[collection_name]


def get_mongo_client(
    db_name, host, port, username, password, ssh_config=None, **kwargs
):
    import pymongo

    if ssh_config is not None:
        client = get_forwarded_mongo_client(
            db_name, username, password, ssh_config, **kwargs
        )
    else:
        client = pymongo.MongoClient(
            host,
            int(port),
            username=username,
            password=password,
            authSource=db_name,
            **kwargs,
        )
    return client


def get_database(db_name, host, port, username, password, **kwargs):
    db = get_mongo_client(db_name, host, port, username, password, **kwargs)[db_name]
    return db


def get_collections_from_mongo_shell_or_pymongo(
    db_name: str, host: str, port: int, username: str, password: str, **kwargs
) -> List[str]:
    """Gets all collections in the database by first using the mongo shell and if that fails uses pymongo.

    Args:
        db_name (str): the name of the database
        host (str): the name of the host
        port (int): the port at which to access the mongodb
        username (str): the username
        password (str): the password

    Returns:
        List[str]: all collections in the database
    """
    import subprocess

    cmd = (
        f"mongo -u '{username}' --authenticationDatabase '{db_name}' {host}:{port}/{db_name} -p {password} "
        "--eval 'db.getCollectionNames().forEach(function(f){print(f)})' --quiet"
    )
    try:
        output = subprocess.check_output(cmd, shell=True, stderr=subprocess.DEVNULL)
        collection_names = output.decode('utf-8').split('\n')
    except (subprocess.CalledProcessError, SyntaxError):
        db = get_database(db_name, host, port, username, password, **kwargs)
        collection_names = db.list_collection_names()
    return [name for name in collection_names if name not in ('fs.chunks', 'fs.files')]


def get_mongodb_config(path=SETTINGS.DATABASE.MONGODB_CONFIG_PATH):
    """Read the MongoDB connection configuration.

    Reads the file at the provided path or otherwise {SETTINGS.DATABASE.MONGODB_CONFIG_PATH} to get
        - database host
        - database port
        - database name
        - username
        - password
        - directConnection (Optional)
        - ssh_config (Optional)

    Default path is $HOME/.config/seml/mongodb.config.

    Config file should be in the format:
    username: <your_username>
    password: <your_password>
    port: <port>
    database: <database_name>
    host: <host>
    directConnection: <bool> (Optional)
    ssh_config: <dict> (Optional)
      ssh_address_or_host: <the url of the jump host>
      ssh_pkey: <the ssh host key>
      ssh_username: <username for jump host>
      retries_max: <number of retries to establish shh tunnel, default 6> (Optional)
      retries_delay: <initial wait time for exponential retry, default 1> (Optional)
      lock_file: <lockfile to avoid establishing ssh tunnel parallely, default `~/seml_ssh.lock`> (Optional)
      lock_timeout: <timeout for aquiring lock, default 30> (Optional)
      ** further arguments passed to `SSHTunnelForwarder` (see https://github.com/pahaz/sshtunnel)

    Returns
    -------
    dict
        Contains the MongoDB config as detailed above.

    """
    import yaml

    access_dict = {}
    config_str = '\nPlease run `seml configure` to provide your credentials.'

    if not path.exists():
        raise MongoDBError(
            f"MongoDB credentials could not be read at '{path}'.{config_str}"
        )

    with open(path) as conf:
        access_dict = yaml.safe_load(conf)

    required_entries = ['username', 'password', 'port', 'host', 'database']
    for entry in required_entries:
        if entry not in access_dict:
            raise MongoDBError(f"No {entry} found in '{path}'.{config_str}")

    db_port = access_dict['port']
    db_name = access_dict['database']
    db_host = access_dict['host']
    db_username = access_dict['username']
    db_password = access_dict['password']
    # False is the default value for PyMongo > 4.0
    db_direct = (
        access_dict['directConnection'] == 'True'
        if 'directConnection' in access_dict
        else False
    )

    cfg = {
        'password': db_password,
        'username': db_username,
        'host': db_host,
        'db_name': db_name,
        'port': db_port,
        'directConnection': db_direct,
    }

    if 'ssh_config' not in access_dict:
        return cfg

    cfg['ssh_config'] = access_dict['ssh_config']
    cfg['ssh_config']['remote_bind_address'] = (db_host, db_port)
    cfg['directConnection'] = True

    return cfg


def build_filter_dict(filter_states, batch_id, filter_dict, sacred_id=None):
    """
    Construct a dictionary to be used for filtering a MongoDB collection.

    Parameters
    ----------
    filter_states: list
        The states to filter for, e.g. ["RUNNING", "PENDING"].
    batch_id: int
        The batch ID to filter.
    filter_dict: dict
        The filter dict passed via the command line. Values in here take precedence over values passed via
        batch_id or filter_states.
    sacred_id: int
        Sacred ID (_id in the database collection) of the experiment. Takes precedence over other filters.

    Returns
    -------
    filter_dict: dict
    """

    if sacred_id is not None:
        return {'_id': sacred_id}

    if filter_dict is None:
        filter_dict = {}

    if filter_states is not None and len(filter_states) > 0:
        if 'status' not in filter_dict:
            filter_dict['status'] = {'$in': filter_states}
        else:
            logging.warning(
                f"'status' was defined in the filter dictionary passed via the command line (-f): "
                f"{filter_dict['status']} AND --status was set to {filter_states}. "
                f"I'm using the value passed via -f."
            )

    if batch_id is not None:
        if 'batch_id' not in filter_dict:
            filter_dict['batch_id'] = batch_id
        else:
            logging.warning(
                f"'batch_id' was defined in the filter dictionary passed via the command line (-f): "
                f"{filter_dict['status']} AND --batch-id was set to {filter_states}. "
                f"I'm using the value passed via -f."
            )
    return filter_dict


def get_max_in_collection(collection, field: str):
    """
    Find the maximum value in the input collection for the input field.
    Parameters
    ----------
    collection
    field

    Returns
    -------
    max_val: the maximum value in the field.
    """
    import pymongo

    ndocs = collection.count_documents({})
    if field == '_id':
        c = collection.find({}, {'_id': 1})
    else:
        c = collection.find({}, {'_id': 1, field: 1})
    b = c.sort(field, pymongo.DESCENDING).limit(1)
    if ndocs != 0:
        b_next = b.next()
        max_val = b_next.get(field)
    else:
        max_val = None

    return max_val


def upload_file(filename, db_collection, batch_id, filetype):
    """
    Upload a source file to the MongoDB.
    Parameters
    ----------
    filename: str
    db_collection: Collection
    batch_id: int
    filetype: str

    Returns
    -------
    file_id: ID of the inserted file, or None if there was an error.
    """
    import gridfs

    db = db_collection.database
    fs = gridfs.GridFS(db)
    try:
        with open(filename, 'rb') as f:
            db_filename = f'file://{db_collection.name}/{batch_id}/{filename}'
            file_id = fs.put(
                f,
                filename=db_filename,
                metadata={
                    'collection_name': db_collection.name,
                    'batch_id': batch_id,
                    'type': filetype,
                },
            )
            return file_id
    except OSError:
        logging.error(f'IOError: could not read {filename}')
    return None


def delete_files(database, file_ids, progress=False):
    import gridfs

    from seml.console import track

    fs = gridfs.GridFS(database)
    it = track(file_ids, disable=not progress)
    for to_delete in it:
        fs.delete(to_delete)


def clean_unreferenced_artifacts(db_collection_name=None, yes=False):
    """
    Delete orphaned artifacts from the database. That is, artifacts that were generated by experiments, but whose
    experiment's database entry has been removed. This leads to storage accumulation, and this function cleans this
    excess storage.
    Parameters
    ----------
    db_collection_name: str, optional
        Database collection to be scanned. If None, all collections will be scanned.
    yes: bool
        Whether to automatically confirm the deletion dialog.
    Returns
    -------
    None
    """
    from seml.console import prompt, track

    all_collections = not bool(db_collection_name)
    if all_collections:
        config = get_mongodb_config()
        db = get_database(**config)
        collection_names = db.list_collection_names()
    else:
        collection = get_collection(db_collection_name)
        db = collection.database
        collection_names = [collection.name]
    collection_names = set(collection_names)
    collection_blacklist = {'fs.chunks', 'fs.files'}
    collection_names = collection_names - collection_blacklist

    referenced_files = set()
    tq = track(collection_names)
    logging.info('Scanning collections for orphaned artifacts...')
    for collection_name in tq:
        collection = db[collection_name]
        experiments = list(
            collection.find(
                {}, {'artifacts': 1, 'experiment.sources': 1, 'source_files': 1}
            )
        )
        for exp in experiments:
            if 'artifacts' in exp:
                try:
                    referenced_files.update({x[1] for x in exp['artifacts']})
                except KeyError:
                    referenced_files.update({x['file_id'] for x in exp['artifacts']})
            if 'experiment' in exp and 'sources' in exp['experiment']:
                referenced_files.update({x[1] for x in exp['experiment']['sources']})
            if 'source_files' in exp:
                referenced_files.update({x[1] for x in exp['source_files']})

    all_files_in_db = list(
        db['fs.files'].find({}, {'_id': 1, 'filename': 1, 'metadata': 1})
    )
    filtered_file_ids = set()
    for file in all_files_in_db:
        if 'filename' in file:
            filename = file['filename']
            file_collection = None
            if (
                filename.startswith('file://')
                and 'metadata' in file
                and file['metadata']
            ):
                # seml-uploaded source
                metadata = file['metadata']
                file_collection = metadata.get('collection_name')
            elif filename.startswith('artifact://'):
                # artifact uploaded by Sacred
                filename = filename[11:]
                file_collection = filename.split('/')[0]
            if (
                file_collection is not None and file_collection in collection_names
            ) or all_collections:
                # only delete files corresponding to collections we want to clean
                filtered_file_ids.add(file['_id'])

    not_referenced_artifacts = filtered_file_ids - referenced_files
    n_delete = len(not_referenced_artifacts)
    if n_delete == 0:
        logging.info('No unreferenced artifacts found.')
        return

    logging.info(
        f'Deleting {n_delete} not referenced artifact{s_if(n_delete)} from database {db.name}. WARNING: This cannot be undone! Artifacts/ files might have been inserted to MongoDB manually or by tools other than seml/ sacred. They will be deleted.'
    )
    if not yes and not prompt('Are you sure? (y/n)', type=bool):
        exit(1)
    logging.info('Deleting not referenced artifacts...')
    delete_files(db, not_referenced_artifacts, progress=True)
    logging.info(
        f'Successfully deleted {n_delete} not referenced artifact{s_if(n_delete)}.'
    )


def update_working_dir(
    db_collection_name: str,
    working_directory: str,
    batch_ids: Optional[List[int]] = None,
):
    """Changes the working directory of experiments in the database.

    Parameters
    ----------
    db_collection_name : str
        The collection to change the working directory in.
    working_directory: str
        The new working directory.
    batch_ids : Optional[List[int]], optional
        Filter on the batch ids, by default None
    """

    collection = get_collection(db_collection_name)
    if batch_ids is not None and len(batch_ids) > 0:
        filter_dict = {'batch_id': {'$in': list(batch_ids)}}
    else:
        filter_dict = {}

    working_directory = str(Path(working_directory).expanduser().resolve())

    update_result = collection.update_many(
        filter_dict,
        {'$set': {'seml.working_dir': working_directory}},
    )
    logging.info(
        f'Updated the working directory of {update_result.modified_count} experiments.'
    )



import logging
import os
import re
import time
from collections import defaultdict
from io import TextIOWrapper
from typing import Dict, List, Optional, Sequence

from seml.commands.manage import detect_duplicates, detect_killed, should_check_killed
from seml.database import (
    build_filter_dict,
    get_collection,
    get_database,
    get_mongodb_config,
)
from seml.experiment.command import (
    get_command_from_exp,
    get_config_overrides,
    get_environment_variables,
    get_shell_command,
)
from seml.settings import SETTINGS
from seml.utils import (
    find_jupyter_host,
    flatten,
    get_from_nested,
    resolve_projection_path_conflicts,
    s_if,
    slice_to_str,
    to_hashable,
    to_slices,
)
from seml.utils.slurm import get_cluster_name, get_slurm_jobs

States = SETTINGS.STATES


def print_fail_trace(
    db_collection_name: str,
    sacred_id: Optional[int],
    filter_states: Optional[List[str]],
    batch_id: Optional[int],
    filter_dict: Optional[Dict],
    projection: Optional[List[str]] = None,
):
    """Convenience function that prints the fail trace of experiments

    Parameters
    ----------
    db_collection_name : str
        Name of the collection to print traces of
    sacred_id : Optional[int]
        Optional filter on the experiment ID
    filter_states : Optional[List[str]]
        Optional filter on the experiment states
    batch_id : Optional[int]
        Optional filter on the experiment batch ID
    filter_dict : Optional[Dict]
        Optional filters
    projection : Optional[List[str]]
        Additional values to print per failed experiment, by default None
    """
    from rich.align import Align
    from rich.console import Group
    from rich.panel import Panel
    from rich.rule import Rule
    from rich.text import Text

    from seml.console import Table, console

    detect_killed(db_collection_name, print_detected=False)
    collection = get_collection(db_collection_name)
    if projection is None:
        projection = []
    mongo_db_projection = resolve_projection_path_conflicts(
        {
            '_id': 1,
            'status': 1,
            'execution.array_id': 1,
            'execution.task_id': 1,
            'fail_trace': 1,
            'seml.description': 1,
            'batch_id': 1,
            **{key: 1 for key in projection},
        }
    )

    if sacred_id is None:
        filter_dict = build_filter_dict(filter_states, batch_id, filter_dict)
        exps = list(collection.find(filter_dict, mongo_db_projection))
    else:
        exps = [collection.find_one({'_id': sacred_id}, mongo_db_projection)]
    for exp in exps:
        exp_id = exp.get('_id')
        status = exp.get('status')
        batch_id = exp.get('batch_id')
        slurm_array_id = exp.get('execution', {}).get('array_id', None)
        slurm_task_id = exp.get('execution', {}).get('task_id', None)
        fail_trace = exp.get('fail_trace', [])
        description = exp.get('seml', {}).get('description', None)
        header = (
            f'Experiment ID {exp_id}, '
            f'Batch ID {batch_id}, '
            f'Status: "{status}", '
            f'Slurm Array-Task id: {slurm_array_id}-{slurm_task_id}'
        )

        renderables = []
        if description is not None:
            text_description = Text()
            text_description.append('Description: ', style='bold magenta')
            text_description.append(description)
            renderables += [text_description, Rule(Text('Fail-Trace', style='bold'))]
        renderables.append(''.join(['\t' + line for line in fail_trace] + []).strip())
        if len(projection) > 0:
            table_projection = Table(show_header=False)
            projection_keys_flat = [
                key
                for key in flatten(exp)
                if any(key.startswith(p) for p in projection)
            ]
            for key in projection_keys_flat:
                table_projection.add_row(key, str(get_from_nested(exp, key)))
            renderables += [
                Rule(Text('Projection', style='bold')),
                Align(table_projection, align='left'),
            ]
        panel = Panel(
            Group(*renderables),
            title=console.render_str(header, highlight=True),
            highlight=True,
            border_style='red',
        )
        console.print(panel)
    logging.info(f'Printed the fail traces of {len(exps)} experiment(s).')


def print_status(
    db_collection_name: str,
    update_status: bool = True,
    projection: Optional[List[str]] = None,
):
    """Prints the status of an experiment collection

    Parameters
    ----------
    db_collection_name : str
        Name of the collection to print status of
    update_status : bool, optional
        Whehter to detect killed experiments, by default True
    projection : Optional[List[str]], optional
        Additional attributes from the MongoDB to print, by default None
    """

    from rich.align import Align
    from rich.table import Column

    from seml.console import Table, console

    collection = get_collection(db_collection_name)

    # Handle status updates
    if update_status:
        detect_killed(db_collection_name, print_detected=False)
    else:
        logging.warning(
            f'Status of {States.RUNNING[0]} experiments may not reflect if they have died or been canceled. Use the `--update-status` flag instead.'
        )

    if projection is None:
        projection = []
    projection = list(resolve_projection_path_conflicts({key: 1 for key in projection}))

    result = collection.aggregate(
        [
            {
                '$group': {
                    '_id': '$status',
                    'ids': {'$addToSet': '$_id'},
                    'batch_ids': {'$addToSet': '$batch_id'},
                    'descriptions': {'$addToSet': '$seml.description'},
                    'count': {'$sum': 1},
                    **{
                        f'projection_{idx}': {'$addToSet': f'${key}'}
                        for idx, key in enumerate(projection)
                    },
                }
            }
        ]
    )
    result = sorted(result, key=lambda x: list(States.keys()).index(x['_id']))
    show_descriptions = any(len(row['descriptions']) > 0 for row in result)
    # Unpack the (nested) projections
    # We keep prefixes encoded as ${id} to preserve the order of the projection keys
    result_projection = []
    for record in result:
        result_projection.append(defaultdict(set))
        for idx, key in enumerate(projection):
            for values in record[f'projection_{idx}']:
                for x, y in flatten({f'${idx}': values}).items():
                    result_projection[-1][x].add(to_hashable(y))
    projection_columns = sorted({k for record in result_projection for k in record})
    # For the column headers, we replace ${id} with the projection key
    columns = []
    for projection_column in projection_columns:
        projection_key_idx = int(
            re.match(r'.*\$([0-9]+)(\..*|$)', projection_column).groups()[0]
        )
        columns.append(
            projection_column.replace(
                f'${projection_key_idx}', projection[projection_key_idx]
            )
        )
    duplicate_experiment_ids = {
        experiment_id
        for dups in detect_duplicates(db_collection_name)
        for experiment_id in dups
    }

    if show_descriptions:
        columns.insert(0, 'Descriptions')
    table = Table(
        Column('Status', justify='left', footer='Total'),
        Column(
            'Count',
            justify='left',
            footer=str(sum(record['count'] for record in result)),
        ),
        Column('Experiment IDs', justify='left'),
        Column('Batch IDs', justify='left'),
        Column('Duplicates', footer=str(len(duplicate_experiment_ids))),
        *[Column(key, justify='left') for key in columns],
        show_header=True,
        show_footer=len(result) > 1,
    )
    for record, record_projection in zip(result, result_projection):
        row = [
            record['_id'],
            str(record['count']),
            ', '.join(map(slice_to_str, to_slices(record['ids']))),
            ', '.join(map(slice_to_str, to_slices(record['batch_ids']))),
            str(len(set(record['ids']) & duplicate_experiment_ids)),
        ]
        if show_descriptions:
            row.append(
                ', '.join(
                    [f'"{description}"' for description in record['descriptions']]
                    if len(record['descriptions']) > 1
                    else record['descriptions']
                )
            )
        row += [
            ', '.join(map(str, record_projection.get(key, {})))
            for key in projection_columns
        ]
        table.add_row(*row)
    console.print(Align(table, align='center'))


def print_collections(
    pattern: str,
    mongodb_config: Optional[Dict] = None,
    progress: bool = False,
    list_empty: bool = False,
    update_status: bool = False,
    print_full_description: bool = False,
):
    """
    Prints a tabular version of multiple collections and their states (without resolving RUNNING experiments that may have been canceled manually).

    Parameters
    ----------
    pattern : str
        The regex collection names have to match against
    mongodb_config : dict or None
        A configuration for the mongodb. If None, the standard config is used.
    progress : bool
        Whether to use a progress bar for fetching
    list_empty : bool
        Whether to list collections that have no documents associated with any state
    update_status : bool
        Whether to update the status of experiments by checking log files. This may take a while.
    print_full_description : bool
        Whether to print full descriptions (wrap-arround) or truncate the descriptions otherwise.
    """
    import pandas as pd
    from rich.align import Align
    from rich.table import Column

    from seml.console import Table, console, track

    # Get the database
    if mongodb_config is None:
        mongodb_config = get_mongodb_config()
    db = get_database(**mongodb_config)
    expression = re.compile(pattern)
    collection_names = [
        name
        for name in db.list_collection_names()
        if name not in ('fs.chunks', 'fs.files') and expression.match(name)
    ]
    # Handle status updates
    if update_status:
        for collection in collection_names:
            detect_killed(collection, print_detected=False)
    else:
        logging.warning(
            f'Status of {States.RUNNING[0]} experiments may not reflect if they have died or been canceled. Use the `--update-status` flag instead.'
        )

    # Count the number of experiments in each state
    name_to_counts = defaultdict(lambda: {state: 0 for state in States.keys()})
    name_to_descriptions = defaultdict(str)
    it = track(collection_names, disable=not progress)

    inv_states = {v: k for k, states in States.items() for v in states}
    show_description = False
    for collection_name in it:
        counts_by_status = db[collection_name].aggregate(
            [
                {
                    '$group': {
                        '_id': '$status',
                        '_count': {'$sum': 1},
                        'description': {'$addToSet': '$seml.description'},
                    }
                }
            ]
        )
        descriptions = db[collection_name].aggregate(
            [{'$group': {'_id': '$seml.description'}}]
        )
        descriptions = [
            result['_id'] for result in descriptions if result['_id'] is not None
        ]
        name_to_counts[collection_name].update(
            {
                inv_states[result['_id']]: result['_count']
                for result in counts_by_status
                if result['_id'] in inv_states
            }
        )
        if len(descriptions) > 1:
            descriptions = [f'"{description}"' for description in descriptions]
        if len(descriptions) > 0:
            show_description = True
        name_to_descriptions[collection_name] = ', '.join(descriptions)

    if len(name_to_counts) == 0:
        logging.info(f'Found no collection matching "{pattern}"!')
        return

    df = pd.DataFrame.from_dict(name_to_counts, dtype=int).transpose()
    # Remove empty collections
    if not list_empty:
        df = df[df.sum(axis=1) > 0]
    # sort rows and columns
    df = df.sort_index()[States.keys()]
    # add a column with the total
    df['Total'] = df.sum(axis=1)

    totals = df.sum(axis=0)
    max_len = max(map(len, collection_names))
    columns = [
        Column('Collection', justify='left', footer='Total', min_width=max_len),
    ] + [
        Column(state.capitalize(), justify='right', footer=str(totals[state]))
        for state in df.columns
    ]
    if show_description:
        columns.append(
            Column(
                'Description(s)',
                justify='left',
                max_width=console.width - max_len - sum(map(len, df.columns)) + 1,
                no_wrap=not print_full_description,
                overflow='ellipsis',
            )
        )
    table = Table(*columns, show_footer=df.shape[0] > 1)
    for collection_name, row in df.iterrows():
        row = [collection_name, *[str(x) for x in row.to_list()]]
        if show_description:
            row.append(name_to_descriptions[collection_name])
        table.add_row(*row)
    # For some reason the table thinks the terminal is larger than it is
    table = Align(table, align='center', width=console.width - max_len + 1)
    console.print(Align(table, align='center'), soft_wrap=True)


def print_duplicates(
    db_collection_name: str,
    filter_states: Optional[List[str]] = None,
    batch_id: Optional[int] = None,
    filter_dict: Optional[Dict] = None,
):
    """Detects and lists duplicate experiment configurations

    Parameters
    ----------
    db_collection_name : str
        The collection to detect duplicates in
    filter_states : Optional[List[str]], optional
        Optional filter on states, by default None
    batch_id : Optional[int], optional
        Optional filter on batch IDs, by default None
    filter_dict : Optional[Dict], optional
        Optional additional user filters, by default None
    """
    from rich.panel import Panel
    from rich.text import Text

    from seml.console import console

    if should_check_killed(filter_states):
        detect_killed(db_collection_name, print_detected=False)
    filter_dict = build_filter_dict(
        filter_states, batch_id, filter_dict, sacred_id=None
    )
    duplicates = detect_duplicates(db_collection_name, filter_dict)
    num_duplicates = sum(map(len, duplicates))
    sorted_duplicates = sorted(list(map(lambda d: tuple(sorted(d)), duplicates)))
    panel = Panel(
        Text.assemble(
            ('Duplicate experiment ID groups: ', 'bold'),
            (', '.join(map(str, sorted_duplicates))),
        ),
        title=console.render_str(
            f'Found {num_duplicates} duplicate experiment configurations ({len(duplicates)} groups)'
        ),
        highlight=True,
        border_style='red',
    )
    console.print(panel)


def print_experiment(
    db_collection_name: str,
    sacred_id: Optional[int] = None,
    filter_states: Optional[List[str]] = None,
    batch_id: Optional[int] = None,
    filter_dict: Optional[Dict] = None,
    projection: Optional[List[str]] = None,
):
    """
    Prints the details of an experiment.

    Parameters
    ----------
    db_collection_name : str
        The collection to print the experiment from
    sacred_id : Optional[int], optional
        The ID of the experiment to print, by default None
    filter_states : Optional[List[str]], optional
        Filter on experiment states, by default None
    batch_id : Optional[int], optional
        Filter on the batch ID of experiments, by default None
    filter_dict : Optional[Dict], optional
        Additional filters, by default None
    projection : Optional[List[str]], optional
        Additional values to print per experiment, by default all are printed
    """
    from rich import print_json

    from seml.console import Heading, console, pause_live_widget

    filter_dict = build_filter_dict(filter_states, batch_id, filter_dict, sacred_id)
    collection = get_collection(db_collection_name)
    if projection is None or len(projection) == 0:
        proj = {}
    else:
        proj = {'_id': 1, 'batch_id': 1, **{p: 1 for p in projection}}
    experiments = list(collection.find(filter_dict, proj))

    if len(experiments) == 0:
        logging.info('No experiment found to print.')
        return

    with pause_live_widget():
        for exp in experiments:
            console.print(Heading(f'Experiment {exp["_id"]} (batch {exp["batch_id"]})'))
            print_json(data=exp, skip_keys=True, default=str)


def print_output(
    db_collection_name: str,
    sacred_id: Optional[int] = None,
    filter_states: Optional[List[str]] = None,
    batch_id: Optional[int] = None,
    filter_dict: Optional[Dict] = None,
    slurm: bool = False,
    head: Optional[int] = None,
    tail: Optional[int] = None,
):
    """
    Prints the output of experiments

    Parameters
    ----------
    db_collection_name : str
        The collection to print the output of
    sacred_id : Optional[int], optional
        The ID of the experiment to print the output of, by default None
    filter_states : Optional[List[str]], optional
        Filter on experiment states, by default None
    batch_id : Optional[int], optional
        Filter on the batch ID of experiments, by default None
    filter_dict : Optional[Dict], optional
        Additional filters, by default None
    slurm : bool, optional
        Whether to print the Slurm output instead of the experiment output, by default False
    head : Optional[int], optional
        The number of lines to print from the beginning of the output, by default None
    tail : Optional[int], optional
        The number of lines to print from the end of the output, by default None
    """
    from seml.console import Heading, console, pause_live_widget

    assert tail is None or head is None, 'Cannot specify both tail and head.'

    filter_dict = build_filter_dict(
        filter_states, batch_id, filter_dict, sacred_id=sacred_id
    )
    collection = get_collection(db_collection_name)
    experiments = collection.find(
        filter_dict,
        {
            'seml.output_file': 1,
            '_id': 1,
            'batch_id': 1,
            'captured_out': 1,
            'execution': 1,
        },
    )

    def print_head(file: TextIOWrapper):
        for i, line in enumerate(file):
            console.print(line[:-1], end=line[-1])
            if head is not None and i >= head:
                break

    def print_tail(file: TextIOWrapper):
        from collections import deque

        last_rows: deque[str] = deque([], maxlen=tail)
        for line in file:
            last_rows.append(line)

        for line in last_rows:
            console.print(line[:-1], end=line[-1])

    def print_all(file: TextIOWrapper):
        for line in file:
            console.print(line[:-1], end=line[-1])

    count = 0
    for exp in experiments:
        count += 1
        console.print(Heading(f'Experiment {exp["_id"]} (batch {exp["batch_id"]})'))
        with pause_live_widget():
            # Select output file
            out_file = exp['seml'].get('output_file')
            if out_file is None or slurm:
                if not slurm:
                    logging.info(
                        f'No experiment output file found for experiment {exp["_id"]}.'
                        'Using Slurm output instead.'
                    )
                if 'slurm_output_file' in exp['execution']:
                    out_file = exp['execution']['slurm_output_file']
                else:
                    logging.error(
                        f'No Slurm output file found for experiment {exp["_id"]}.'
                    )
                    continue
            # Actually read
            try:
                with open(out_file, newline='', errors='replace') as f:
                    if head:
                        print_head(f)
                    elif tail:
                        print_tail(f)
                    else:
                        print_all(f)
                console.print()  # new line
            except OSError:
                logging.info(f'File {out_file} could not be read.')
                if 'captured_out' in exp and exp['captured_out']:
                    logging.info('Captured output from DB:')
                    console.print(exp['captured_out'])
                else:
                    logging.error('No output available.')

    if count == 0:
        logging.info('No experiments found.')


def generate_queue_table(
    db,
    job_ids: Optional[List[str]],
    filter_states: Optional[List[str]],
    filter_by_user: bool = True,
):
    """
    Generates a table of the SEML collections of Slurm jobs.

    Parameters
    ----------
    job_ids : List[str]
        The job IDs to check
    filter_by_user : bool, optional
        Whether to only check jobs by the current user, by default True.

    Returns
    -------
    Align
        The table of the SEML collections of Slurm jobs.
    """
    from rich.align import Align

    from seml.console import Table

    if job_ids:
        job_infos = get_slurm_jobs(*job_ids)
    else:
        job_infos = get_slurm_jobs()

    # Find the collections
    all_collections = set(db.list_collection_names())

    collection_to_jobs = defaultdict(list)
    states = set()
    collections = set()
    for job in job_infos:
        state = job['JobState']
        if filter_states is not None and state not in filter_states:
            continue

        user_id = job.get('UserId', '').split('(')[0]
        if filter_by_user and user_id != os.environ['USER']:
            continue

        collection = job.get('Comment', None)
        if not (collection and collection in all_collections):
            collection = 'No collection found'
            if job['JobName'] == 'jupyter':
                collection = 'Jupyter'
                url, known_host = find_jupyter_host(job['StdOut'], False)
                if known_host is not None:
                    collection = f'Jupyter ({url})'

        collection_to_jobs[(collection, state)].append(job)
        states.add(job['JobState'])
        collections.add(collection)

    # Print the collections
    sorted_states = sorted(states)
    table = Table(
        'Collection',
        *map(str.title, sorted_states),
        show_header=True,
    )
    cluster_name = get_cluster_name()

    def format_job(job_info, db_col_name):
        if job_info is None:
            return ''
        nodelist = job_info['NodeList']
        array_id = job_info.get('ArrayJobId', job_info['JobId'])
        task_id = job_info.get('ArrayTaskId', None)
        if task_id:
            if any(x in task_id for x in ',-'):
                task_id = f'[{task_id}]'
            job_id = f'{array_id}_{task_id}'
        else:
            job_id = array_id
        if nodelist:
            # Running
            collection = get_collection(db_col_name)
            experiments = collection.find(
                {
                    'execution.cluster': cluster_name,
                    'execution.array_id': int(array_id),
                    'execution.task_id': int(task_id),
                },
                {'_id': 1},
            )
            ids = [exp['_id'] for exp in experiments]
            if len(ids) > 0:
                return f"{job_id} ({job_info['RunTime']}, {nodelist}, Id{s_if(len(ids))}: {'|'.join(map(str, ids))})"
            else:
                return f"{job_id} ({job_info['RunTime']}, {nodelist})"
        else:
            return f"{job_id} ({job_info.get('Reason', '')})"

    for col in sorted(collections):
        row = [col]
        for state in sorted_states:
            jobs = collection_to_jobs[(col, state)]
            row.append('\n'.join(map(format_job, jobs, [col] * len(jobs))))
        table.add_row(*row)

    if len(collections) == 0:
        return Align('No jobs found.', align='center')
    return Align(table, align='center')


def print_queue(
    job_ids: Optional[Sequence[str]],
    filter_states: Optional[Sequence[str]],
    filter_by_user: bool,
    watch: bool,
):
    """
    Prints the SEML collections of Slurm jobs.

    Parameters
    ----------
    job_ids : Optional[Sequence[str]], optional
        The job IDs to check, by default None (None meaning all jobs)
    filter_by_user : bool, optional
        Whether to only check jobs by the current user, by default True
    """
    from rich.live import Live

    from seml.console import console, pause_live_widget

    mongodb_config = get_mongodb_config()
    db = get_database(**mongodb_config)

    def generate_table_fn():
        return generate_queue_table(db, job_ids, filter_states, filter_by_user)

    table = generate_table_fn()
    if watch:
        console.clear()
        with pause_live_widget():
            with Live(table, refresh_per_second=0.5) as live:
                while True:
                    time.sleep(2)
                    live.update(generate_table_fn())
    else:
        console.print(table)


def print_command(
    db_collection_name: str,
    sacred_id: Optional[int],
    batch_id: Optional[int],
    filter_states: List[str],
    filter_dict: Optional[Dict],
    num_exps: int,
    worker_gpus: Optional[str] = None,
    worker_cpus: Optional[int] = None,
    worker_environment_vars: Optional[Dict] = None,
    unresolved: bool = False,
    resolve_interpolations: bool = True,
):
    import rich

    from seml.console import Heading, console

    collection = get_collection(db_collection_name)

    filter_dict = build_filter_dict(filter_states, batch_id, filter_dict, sacred_id)

    env_dict = get_environment_variables(
        worker_gpus, worker_cpus, worker_environment_vars
    )

    orig_level = logging.root.level
    logging.root.setLevel(logging.NOTSET)

    exps_list = list(collection.find(filter_dict, limit=num_exps))
    if len(exps_list) == 0:
        return

    exp = exps_list[0]
    _, exe, config = get_command_from_exp(
        exp,
        collection.name,
        verbose=logging.root.level <= logging.VERBOSE,
        unobserved=True,
        post_mortem=False,
        unresolved=unresolved,
        resolve_interpolations=resolve_interpolations,
    )
    _, exe, vscode_config = get_command_from_exp(
        exp,
        collection.name,
        verbose=logging.root.level <= logging.VERBOSE,
        unobserved=True,
        post_mortem=False,
        use_json=True,
        unresolved=unresolved,
        resolve_interpolations=resolve_interpolations,
    )
    env = exp['seml'].get('conda_environment')

    console.print(Heading('First experiment'))
    logging.info(f'Executable: {exe}')
    if env is not None:
        logging.info(f'Anaconda environment: {env}')

    console.print(Heading('Arguments for VS Code debugger'))
    rich.print_json(data=['with', '--debug'] + vscode_config)
    console.print(Heading('Arguments for PyCharm debugger'))
    print('with --debug ' + get_config_overrides(config))

    console.print(Heading('Command for post-mortem debugging'))
    interpreter, exe, config = get_command_from_exp(
        exps_list[0],
        collection.name,
        verbose=logging.root.level <= logging.VERBOSE,
        unobserved=True,
        post_mortem=True,
        unresolved=unresolved,
        resolve_interpolations=resolve_interpolations,
    )
    print(get_shell_command(interpreter, exe, config, env=env_dict))

    console.print(Heading('Command for remote debugging'))
    interpreter, exe, config = get_command_from_exp(
        exps_list[0],
        collection.name,
        verbose=logging.root.level <= logging.VERBOSE,
        unobserved=True,
        debug_server=True,
        print_info=False,
        unresolved=unresolved,
        resolve_interpolations=resolve_interpolations,
    )
    print(get_shell_command(interpreter, exe, config, env=env_dict))

    console.print(Heading('All raw commands'))
    logging.root.setLevel(orig_level)
    for exp in exps_list:
        interpreter, exe, config = get_command_from_exp(
            exp,
            collection.name,
            verbose=logging.root.level <= logging.VERBOSE,
            unresolved=unresolved,
            resolve_interpolations=resolve_interpolations,
        )
        print(get_shell_command(interpreter, exe, config, env=env_dict))



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .experiment import Experiment  # noqa
else:

    def Experiment(*args, **kwargs):
        # THis proxy class is used to avoid importing the real thing
        from .experiment import Experiment as SemlExperiment

        return SemlExperiment(*args, **kwargs)



import copy
import datetime
import logging
import os
from typing import TYPE_CHECKING, Dict, List, Optional

from seml.database import get_collection, get_max_in_collection
from seml.experiment.config import (
    check_config,
    config_get_exclude_keys,
    generate_configs,
    generate_named_configs,
    read_config,
    remove_prepended_dashes,
    resolve_configs,
    resolve_interpolations,
)
from seml.experiment.description import resolve_description
from seml.experiment.sources import get_git_info, upload_sources
from seml.settings import SETTINGS
from seml.utils import (
    Hashabledict,
    flatten,
    make_hash,
    merge_dicts,
    remove_keys_from_nested,
    s_if,
    unflatten,
)
from seml.utils.errors import ConfigError

if TYPE_CHECKING:
    import pymongo

States = SETTINGS.STATES


def filter_experiments(
    collection: 'pymongo.collection.Collection',
    documents: List[Dict],
    use_hash: bool = True,
):
    """Check database collection for already present entries.

    Check the database collection for experiments that have the same configuration.
    Remove the corresponding entries from the input list of configurations to prevent
    re-running the experiments.

    Parameters
    ----------
    collection: pymongo.collection.Collection
        The MongoDB collection containing the experiments.
    documents: List[Dict]
        The documents to filter.
    use_hash : bool
        Whether to use hashes (faster)

    Returns
    -------
    filtered_configs: list of dicts
        No longer contains configurations that are already in the database collection.

    """
    filtered_documents = []
    for document in documents:
        if use_hash:
            lookup_result = collection.find_one(
                {'config_hash': document['config_hash']}
            )
        else:
            lookup_dict = flatten(
                {
                    'config': remove_keys_from_nested(
                        document['config'], document['config_unresolved']
                    )
                }
            )
            lookup_result = collection.find_one(unflatten(lookup_dict))
        if lookup_result is None:
            filtered_documents.append(document)
    return filtered_documents


def add_configs(
    collection: 'pymongo.collection.Collection',
    documents: List[Dict],
    description: Optional[str] = None,
    resolve_descriptions: bool = True,
):
    """Put the input configurations into the database.

    Parameters
    ----------
    collection: pymongo.collection.Collection
        The MongoDB collection containing the experiments.
    documents : List[Dict]
        The documents to add.
    description : Optional[str], optional
        Optional description for the experiments, by default None
    resolve_descriptions : bool, optional
        Whether to use omegaconf to resolve descriptions.
    """

    if len(documents) == 0:
        return

    start_id = get_max_in_collection(collection, '_id')
    if start_id is None:
        start_id = 1
    else:
        start_id = start_id + 1

    logging.info(
        f"Adding {len(documents)} configs to the database (batch-ID {documents[0]['batch_id']})."
    )

    documents = [
        {
            **document,
            **{
                '_id': start_id + idx,
                'status': States.STAGED[0],
                'add_time': datetime.datetime.utcnow(),
            },
        }
        for idx, document in enumerate(documents)
    ]
    if description is not None:
        for db_dict in documents:
            db_dict['seml']['description'] = description
    if resolve_descriptions:
        for db_dict in documents:
            if 'description' in db_dict['seml']:
                db_dict['seml']['description'] = resolve_description(
                    db_dict['seml']['description'], db_dict
                )

    collection.insert_many(documents)


def add_config_files(
    db_collection_name: str,
    config_files: List[str],
    force_duplicates: bool = False,
    overwrite_params: Optional[Dict] = None,
    no_hash: bool = False,
    no_sanity_check: bool = False,
    no_code_checkpoint: bool = False,
    description: Optional[str] = None,
    resolve_descriptions: bool = True,
):
    """Adds configuration files to the MongoDB

    Parameters
    ----------
    db_collection_name : str
        The collection to which to add
    config_files : List[str]
        A list of paths to configuration files (YAML) to add
    force_duplicates : bool, optional
        Whether to force adding configurations even if they are already in the MongoDB, by default False
    overwrite_params : Optional[Dict], optional
        Optional flat dict to override configuration values, by default None
    no_hash : bool, optional
        Whether to skip duplicate detection by computing hashes, which may result in slowdowns, by default False
    no_sanity_check : bool, optional
        Whether to skip feeding configuration values into the sacred experiment to detect unsupported or missing keys, by default False
    no_code_checkpoint : bool, optional
        Whether to not base the experiments on a copy of the current codebase, by default False
    description : Optional[str], optional
        Optional description for the experiments, by default None
    resolve_descriptions : bool, optional
        Whether to use omegaconf to resolve experiment descriptions.
    """
    config_files = [os.path.abspath(file) for file in config_files]
    for config_file in config_files:
        add_config_file(
            db_collection_name,
            config_file,
            force_duplicates,
            overwrite_params,
            no_hash,
            no_sanity_check,
            no_code_checkpoint,
            description=description,
            resolve_descriptions=resolve_descriptions,
        )


def assemble_slurm_config_dict(experiment_slurm_config: dict):
    """
    Realize inheritance for the slurm configuration, with the following relationship:
    Default -> Template -> Experiment

    Parameters
    ----------
    experiment_slurm_config: The slurm experiment configuration as returned by the function read_config

    Returns
    -------
    slurm_config

    """
    # Rename
    slurm_config = experiment_slurm_config
    # Assemble the Slurm config:
    # Basis config is the default config. This can be overridden by the sbatch_options_template.
    # And this in turn can be overridden by the sbatch config defined in the experiment .yaml file.
    slurm_config_base = copy.deepcopy(SETTINGS.SLURM_DEFAULT)

    # Check for and use sbatch options template
    sbatch_options_template = slurm_config.get('sbatch_options_template', None)
    if sbatch_options_template is not None:
        if sbatch_options_template not in SETTINGS.SBATCH_OPTIONS_TEMPLATES:
            raise ConfigError(
                f"sbatch options template '{sbatch_options_template}' not found in settings.py."
            )
        slurm_config_base['sbatch_options'] = merge_dicts(
            slurm_config_base['sbatch_options'],
            SETTINGS.SBATCH_OPTIONS_TEMPLATES[sbatch_options_template],
        )

    # Integrate experiment specific config
    slurm_config = merge_dicts(slurm_config_base, slurm_config)

    slurm_config['sbatch_options'] = remove_prepended_dashes(
        slurm_config['sbatch_options']
    )
    return slurm_config


def add_config_file(
    db_collection_name: str,
    config_file: str,
    force_duplicates: bool = False,
    overwrite_params: Optional[Dict] = None,
    no_hash: bool = False,
    no_sanity_check: bool = False,
    no_code_checkpoint: bool = False,
    description: Optional[str] = None,
    resolve_descriptions: bool = True,
):
    """Adds configuration files to the MongoDB

    Parameters
    ----------
    db_collection_name : str
        The collection to which to add
    config_files : str
        A path to configuration file (YAML) to add
    force_duplicates : bool, optional
        Whether to force adding configurations even if they are already in the MongoDB, by default False
    overwrite_params : Optional[Dict], optional
        Optional flat dict to override configuration values, by default None
    no_hash : bool, optional
        Whether to skip duplicate detection by computing hashes, which may result in slowdowns, by default False
    no_sanity_check : bool, optional
        Whether to skip feeding configuration values into the sacred experiment to detect unsupported or missing keys, by default False
    no_code_checkpoint : bool, optional
        Whether to not base the experiments on a copy of the current codebase, by default False
    description : Optional[str], optional
        Optional description for the experiments, by default None
    resolve_descriptions : bool, optional
        Whether to use omegaconf to resolve descriptions
    """

    collection = get_collection(db_collection_name)
    seml_config, slurm_configs, experiment_config = read_config(config_file)

    # Use current Anaconda environment if not specified
    if 'conda_environment' not in seml_config:
        seml_config['conda_environment'] = os.environ.get('CONDA_DEFAULT_ENV')

    path, commit, dirty = get_git_info(
        seml_config['executable'], seml_config['working_dir']
    )
    git_info = None
    if path is not None:
        git_info = {'path': path, 'commit': commit, 'dirty': dirty}

    batch_id = get_max_in_collection(collection, 'batch_id')
    if batch_id is None:
        batch_id = 1
    else:
        batch_id = batch_id + 1

    # Assemble the Slurm config:
    slurm_configs = list(map(assemble_slurm_config_dict, slurm_configs))
    if len(slurm_configs) == 0:
        raise ConfigError('No slurm configuration found.')

    configs_unresolved = generate_configs(
        experiment_config, overwrite_params=overwrite_params
    )
    configs, named_configs = generate_named_configs(configs_unresolved)
    configs = resolve_configs(
        seml_config['executable'],
        seml_config['conda_environment'],
        configs,
        named_configs,
        seml_config['working_dir'],
    )

    # Create documents that can be interpolated
    documents = [
        {
            **resolve_interpolations(
                {
                    'seml': seml_config,
                    'slurm': slurm_configs,
                    'git': git_info,
                    'batch_id': batch_id,  # needs to be determined now for source file uploading
                    'config': config,
                    'config_unresolved': config_unresolved,
                }
            ),
            'config_unresolved': config_unresolved,
        }
        for config, config_unresolved in zip(configs, configs_unresolved)
    ]

    # Upload source files: This also determines the batch_id
    if seml_config['use_uploaded_sources'] and not no_code_checkpoint:
        seml_config['source_files'] = upload_sources(seml_config, collection, batch_id)
    del seml_config['use_uploaded_sources']
    documents = [{**document, **{'seml': seml_config}} for document in documents]

    if not no_sanity_check:
        # Sanity checking uses the resolved values (after considering named configs)
        check_config(
            seml_config['executable'],
            seml_config['conda_environment'],
            [document['config'] for document in documents],
            seml_config['working_dir'],
        )

    use_hash = not no_hash
    for document in documents:
        document['config_hash'] = make_hash(
            document['config'],
            config_get_exclude_keys(document['config'], document['config_unresolved']),
        )

    if not force_duplicates:
        len_before = len(documents)

        # First, check for duplicates withing the experiment configurations from the file.
        if not use_hash:
            # slow duplicate detection without hashes
            unique_documents, unique_keys = [], set()
            for document in documents:
                key = Hashabledict(
                    **remove_keys_from_nested(
                        document['config'],
                        config_get_exclude_keys(
                            document['config'], document['config_unresolved']
                        ),
                    )
                )
                if key not in unique_keys:
                    unique_documents.append(document)
                    unique_keys.add(key)
            documents = unique_documents
        else:
            # fast duplicate detection using hashing.
            documents_dict = {
                document['config_hash']: document for document in documents
            }
            documents = list(documents_dict.values())

        len_after_deduplication = len(documents)
        # Now, check for duplicate configurations in the database.
        documents = filter_experiments(collection, documents, use_hash=use_hash)
        len_after = len(documents)
        if len_after_deduplication != len_before:
            logging.info(
                f'{len_before - len_after_deduplication} of {len_before} experiment{s_if(len_before)} were '
                f'duplicates. Adding only the {len_after_deduplication} unique configurations.'
            )
        if len_after != len_after_deduplication:
            logging.info(
                f'{len_after_deduplication - len_after} of {len_after_deduplication} '
                f'experiment{s_if(len_before)} were already found in the database. They were not added again.'
            )

    # Create an index on the config hash. If the index is already present, this simply does nothing.
    collection.create_index('config_hash')
    # Add the configurations to the database with STAGED status.
    if len(configs) > 0:
        add_configs(
            collection,
            documents,
            description=description,
            resolve_descriptions=resolve_descriptions,
        )



import logging
import os
from contextlib import contextmanager
from pathlib import Path
from typing import List, Optional, Union

from seml.settings import SETTINGS


def init_project(
    directory: Union[str, Path] = '.',
    project_name: Optional[str] = None,
    user_name: Optional[str] = None,
    user_mail: Optional[str] = None,
    template: str = 'default',
    git_remote: Optional[str] = None,
    git_commit: Optional[str] = None,
    yes: bool = False,
):
    """
    Initialize a new project in the given directory.

    Parameters
    ----------
    directory : Union[str, Path]
        The directory to initialize the project in.
    project_name : Optional[str]
        The name of the project. If not given, the name of the directory is used.
    user_name : Optional[str]
        The name of the user. If not given, the environment variable USER is used.
    user_mail : Optional[str]
        The email of the user. If not given, ''.
    template : str
        The template to use for the project.
    git_repo : Optional[str]
        The URL of the git repository to use.
    git_commit : Optional[str]
        The commit to use.
    git_branch : Optional[str]
        The branch to use.
    yes : bool
        If True, no confirmation is asked before initializing the project.
    """
    from gitignore_parser import parse_gitignore

    from seml.console import prompt

    if directory is None:
        directory = Path()
    directory = Path(directory).absolute()

    # Ensure that the directory exists
    if not directory.exists():
        directory.mkdir(parents=True)

    # Ensure that its state is okay
    if any(directory.glob('**/*')) and not yes:
        if not prompt(
            f'Directory "{directory}" is not empty. Are you sure you want to initialize a new project here? (y/n)',
            type=bool,
        ):
            exit(1)

    with checkout_template_repo(git_remote, git_commit) as tmp_dir:
        template_path = tmp_dir / 'templates' / template
        if not template_path.exists():
            logging.error(f'Template "{template}" does not exist')
            exit(1)

        logging.info(
            f'Initializing project in "{directory}" using template "{template}".'
        )

        if project_name is None:
            project_name = directory.name
        if user_name is None:
            user_name = os.getenv('USER', os.getenv('USERNAME', 'user'))
        if user_mail is None:
            user_mail = 'my@mail.com'
        format_map = dict(
            project_name=project_name, user_name=user_name, user_mail=user_mail
        )

        gitignore_path = template_path / '.gitignore'
        if gitignore_path.exists():
            ignore_file = parse_gitignore(gitignore_path)
        else:

            def ignore_file(file_path: str):
                return False

        # Copy files one-by-one
        for src in template_path.glob('**/*'):
            # skip files ignored by .gitignore
            if ignore_file(str(src)):
                continue
            # construct destination
            file_name = src.relative_to(template_path)
            target_file_name = Path(str(file_name).format_map(format_map))
            dst = directory / target_file_name
            # Create directories
            if src.is_dir():
                if not dst.exists():
                    dst.mkdir()
            elif not dst.exists():
                # For templates fill in variables
                if src.suffix.endswith('.template'):
                    dst = dst.with_suffix(src.suffix.replace('.template', ''))
                    dst.write_text(src.read_text().format_map(format_map))
                else:
                    # Other files copy directly
                    dst.write_bytes(src.read_bytes())
    logging.info('Project initialized successfully')


@contextmanager
def checkout_template_repo(
    git_remote: Optional[str] = None, git_commit: Optional[str] = None
):
    """
    Context manager to clone the template repository. The cloned repository
    is deleted after the context is left.

    Parameters
    ----------
    git_remote : Optional[str]
        The git remote to use.
    git_commit : Optional[str]
        The git commit to use.
    """
    import tempfile

    from git import Repo

    if git_remote is None:
        git_remote = SETTINGS.TEMPLATE_REMOTE

    with tempfile.TemporaryDirectory(dir=SETTINGS.TMP_DIRECTORY) as tmp_dir:
        try:
            repo = Repo.clone_from(git_remote, tmp_dir)
            if git_commit is not None:
                repo.head.reference = repo.commit(git_commit)
                repo.head.reset(index=True, working_tree=True)
        except Exception as e:
            logging.error(
                f'Failed to clone git repository "{git_remote}" to "{tmp_dir}"'
            )
            logging.error(e)
            exit(1)
        yield Path(repo.working_dir)


def get_available_templates(
    git_remote: Optional[str] = None, git_commit: Optional[str] = None
) -> List[str]:
    """
    Return a list of available templates.

    Parameters
    ----------
    git_remote : Optional[str]
        The git remote to use.
    git_commit : Optional[str]
        The git commit to use.

    Returns
    -------
    List[str]
        A list of available templates.
    """
    with checkout_template_repo(git_remote, git_commit) as repo:
        return [template.name for template in (repo / 'templates').iterdir()]


def print_available_templates(
    git_remote: Optional[str] = None, git_commit: Optional[str] = None
):
    """
    Print the available templates.

    Parameters
    ----------
    git_remote : Optional[str]
        The git remote to use.
    git_commit : Optional[str]
        The git commit to use.
    """
    result = 'Available templates:'
    for template in get_available_templates(git_remote, git_commit):
        result += f'\n  - {template}'
    logging.info(result)



# We need the custom json encoding for vscode due to https://github.com/microsoft/vscode/issues/91578
# Once this bug has been fixed we should only rely on `repr` and remove this file.
import json


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        import numpy as np
        from bson import json_util

        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        # we default to json_util.default as it can deal with more data types.
        return json_util.default(obj)


class PythonEncoder(json.JSONEncoder):
    def iterencode(self, o, _one_shot=False):
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = json.encoder.encode_basestring_ascii
        else:
            _encoder = json.encoder.encode_basestring

        def floatstr(
            o,
            allow_nan=self.allow_nan,
            _repr=float.__repr__,
            _inf=json.encoder.INFINITY,
            _neginf=-json.encoder.INFINITY,
        ):
            # Check for specials.  Note that this type of test is processor
            # and/or platform-specific, so do tests which don't depend on the
            # internals.

            if o != o:
                text = 'nan'
            elif o == _inf:
                text = 'inf'
            elif o == _neginf:
                text = '-inf'
            else:
                return _repr(o)

            if not allow_nan:
                raise ValueError(
                    'Out of range float values are not JSON compliant: ' + repr(o)
                )

            return text

        _iterencode = _make_iterencode(
            markers,
            self.default,
            _encoder,
            self.indent,
            floatstr,
            self.key_separator,
            self.item_separator,
            self.sort_keys,
            self.skipkeys,
            _one_shot,
        )
        return _iterencode(o, 0)


def _make_iterencode(
    markers,
    _default,
    _encoder,
    _indent,
    _floatstr,
    _key_separator,
    _item_separator,
    _sort_keys,
    _skipkeys,
    _one_shot,
    ## HACK: hand-optimized bytecode; turn globals into locals
    ValueError=ValueError,
    dict=dict,
    float=float,
    id=id,
    int=int,
    isinstance=isinstance,
    list=list,
    str=str,
    tuple=tuple,
    _intstr=int.__repr__,
):
    if _indent is not None and not isinstance(_indent, str):
        _indent = ' ' * _indent

    def _iterencode_list(lst, _current_indent_level):
        if not lst:
            yield '[]'
            return
        if markers is not None:
            markerid = id(lst)
            if markerid in markers:
                raise ValueError('Circular reference detected')
            markers[markerid] = lst
        buf = '['
        if _indent is not None:
            _current_indent_level += 1
            newline_indent = '\n' + _indent * _current_indent_level
            separator = _item_separator + newline_indent
            buf += newline_indent
        else:
            newline_indent = None
            separator = _item_separator
        first = True
        for value in lst:
            if first:
                first = False
            else:
                buf = separator
            if isinstance(value, str):
                yield buf + _encoder(value)
            elif value is None:
                yield buf + 'None'
            elif value is True:
                yield buf + 'True'
            elif value is False:
                yield buf + 'False'
            elif isinstance(value, int):
                # Subclasses of int/float may override __repr__, but we still
                # want to encode them as integers/floats in JSON. One example
                # within the standard library is IntEnum.
                yield buf + _intstr(value)
            elif isinstance(value, float):
                # see comment above for int
                yield buf + _floatstr(value)
            else:
                yield buf
                if isinstance(value, (list, tuple)):
                    chunks = _iterencode_list(value, _current_indent_level)
                elif isinstance(value, dict):
                    chunks = _iterencode_dict(value, _current_indent_level)
                else:
                    chunks = _iterencode(value, _current_indent_level)
                yield from chunks
        if newline_indent is not None:
            _current_indent_level -= 1
            yield '\n' + _indent * _current_indent_level
        yield ']'
        if markers is not None:
            del markers[markerid]

    def _iterencode_dict(dct, _current_indent_level):
        if not dct:
            yield '{}'
            return
        if markers is not None:
            markerid = id(dct)
            if markerid in markers:
                raise ValueError('Circular reference detected')
            markers[markerid] = dct
        yield '{'
        if _indent is not None:
            _current_indent_level += 1
            newline_indent = '\n' + _indent * _current_indent_level
            item_separator = _item_separator + newline_indent
            yield newline_indent
        else:
            newline_indent = None
            item_separator = _item_separator
        first = True
        if _sort_keys:
            items = sorted(dct.items())
        else:
            items = dct.items()
        for key, value in items:
            if isinstance(key, str):
                pass
            # JavaScript is weakly typed for these, so it makes sense to
            # also allow them.  Many encoders seem to do something like this.
            elif isinstance(key, float):
                # see comment for int/float in _make_iterencode
                key = _floatstr(key)
            elif key is True:
                key = 'True'
            elif key is False:
                key = 'False'
            elif key is None:
                key = 'None'
            elif isinstance(key, int):
                # see comment for int/float in _make_iterencode
                key = _intstr(key)
            elif _skipkeys:
                continue
            else:
                raise TypeError(
                    f'keys must be str, int, float, bool or None, '
                    f'not {key.__class__.__name__}'
                )
            if first:
                first = False
            else:
                yield item_separator
            yield _encoder(key)
            yield _key_separator
            if isinstance(value, str):
                yield _encoder(value)
            elif value is None:
                yield 'None'
            elif value is True:
                yield 'True'
            elif value is False:
                yield 'False'
            elif isinstance(value, int):
                # see comment for int/float in _make_iterencode
                yield _intstr(value)
            elif isinstance(value, float):
                # see comment for int/float in _make_iterencode
                yield _floatstr(value)
            else:
                if isinstance(value, (list, tuple)):
                    chunks = _iterencode_list(value, _current_indent_level)
                elif isinstance(value, dict):
                    chunks = _iterencode_dict(value, _current_indent_level)
                else:
                    chunks = _iterencode(value, _current_indent_level)
                yield from chunks
        if newline_indent is not None:
            _current_indent_level -= 1
            yield '\n' + _indent * _current_indent_level
        yield '}'
        if markers is not None:
            del markers[markerid]

    def _iterencode(o, _current_indent_level):
        if isinstance(o, str):
            yield _encoder(o)
        elif o is None:
            yield 'None'
        elif o is True:
            yield 'True'
        elif o is False:
            yield 'False'
        elif isinstance(o, int):
            # see comment for int/float in _make_iterencode
            yield _intstr(o)
        elif isinstance(o, float):
            # see comment for int/float in _make_iterencode
            yield _floatstr(o)
        elif isinstance(o, (list, tuple)):
            yield from _iterencode_list(o, _current_indent_level)
        elif isinstance(o, dict):
            yield from _iterencode_dict(o, _current_indent_level)
        else:
            if markers is not None:
                markerid = id(o)
                if markerid in markers:
                    raise ValueError('Circular reference detected')
                markers[markerid] = o
            o = _default(o)
            yield from _iterencode(o, _current_indent_level)
            if markers is not None:
                del markers[markerid]

    return _iterencode



import itertools
import random
import uuid
from typing import DefaultDict

from seml.utils import unflatten
from seml.utils.errors import ConfigError


def sample_random_configs(random_config, samples=1, seed=None):
    """
    Sample random configurations from the specified search space.

    Parameters
    ----------
    random_config: dict
        dict where each key is a parameter and the value defines how the random sample is drawn. The samples will be
        drawn using the function sample_parameter.
    samples: int
        The number of samples to draw per parameter
    seed: int or None
        The seed to use when drawing the parameter value. Defaults to None.

    Returns
    -------
    random_configurations: list of dicts
        List of dicts, where each dict gives a value for all parameters defined in the input random_config dict.

    """

    if len(random_config) == 0:
        return [{}]

    rdm_keys = [k for k in random_config.keys() if k not in ['samples', 'seed']]
    random_config = {k: random_config[k] for k in rdm_keys}
    random_parameter_dicts = unflatten(random_config, levels=-1)
    random_samples = [
        sample_parameter(random_parameter_dicts[k], samples, seed, parent_key=k)
        for k in random_parameter_dicts.keys()
    ]
    random_samples = dict([sub for item in random_samples for sub in item])
    random_configurations = [
        {k: v[ix] for k, v in random_samples.items()} for ix in range(samples)
    ]

    return random_configurations


def sample_parameter(parameter, samples, seed=None, parent_key=''):
    """
    Generate random samples from the specified parameter.

    The parameter types are inspired from https://github.com/hyperopt/hyperopt/wiki/FMin. When implementing new types,
    please make them compatible with the hyperopt nomenclature so that we can switch to hyperopt at some point.

    Parameters
    ----------
    parameter: dict
        Defines the type of parameter. Dict must include the key "type" that defines how the parameter will be sampled.
        Supported types are
            - choice: Randomly samples <samples> entries (with replacement) from the list in parameter['options']
            - uniform: Uniformly samples between 'min' and 'max' as specified in the parameter dict.
            - loguniform:  Uniformly samples in log space between 'min' and 'max' as specified in the parameter dict.
            - randint: Randomly samples integers between 'min' (included) and 'max' (excluded).
    samples: int
        Number of samples to draw for the parameter.
    seed: int
        The seed to use when drawing the parameter value. Defaults to None.
    parent_key: str
        The key to prepend the parameter name with. Used for nested parameters, where we here create a flattened version
        where e.g. {'a': {'b': 11}, 'c': 3} becomes {'a.b': 11, 'c': 3}

    Returns
    -------
    return_items: tuple(str, np.array or list)
        tuple of the parameter name and a 1-D list/array of the samples drawn for the parameter.

    """
    import numpy as np

    if 'type' not in parameter:
        raise ConfigError(f'No type found in parameter {parameter}')
    return_items = []
    allowed_keys = ['seed', 'type']
    if 'seed' in parameter:
        np.random.seed(parameter['seed'])
    elif seed is not None:
        np.random.seed(seed)

    param_type = parameter['type']

    if param_type == 'choice':
        choices = parameter['options']
        allowed_keys.append('options')
        sampled_values = [random.choice(choices) for _ in range(samples)]
        return_items.append((parent_key, sampled_values))

    elif param_type == 'uniform':
        min_val = parameter['min']
        max_val = parameter['max']
        allowed_keys.extend(['min', 'max'])
        sampled_values = np.random.uniform(min_val, max_val, samples)
        return_items.append((parent_key, sampled_values))

    elif param_type == 'loguniform':
        if parameter['min'] <= 0:
            raise ConfigError('Cannot take log of values <= 0')
        min_val = np.log(parameter['min'])
        max_val = np.log(parameter['max'])
        allowed_keys.extend(['min', 'max'])
        sampled_values = np.exp(np.random.uniform(min_val, max_val, samples))
        return_items.append((parent_key, sampled_values))

    elif param_type == 'randint':
        min_val = int(parameter['min'])
        max_val = int(parameter['max'])
        allowed_keys.extend(['min', 'max'])
        sampled_values = np.random.randint(min_val, max_val, samples)
        return_items.append((parent_key, sampled_values))

    elif param_type == 'randint_unique':
        min_val = int(parameter['min'])
        max_val = int(parameter['max'])
        allowed_keys.extend(['min', 'max'])
        sampled_values = np.random.choice(
            np.arange(min_val, max_val), samples, replace=False
        )
        return_items.append((parent_key, sampled_values))

    elif param_type == 'parameter_collection':
        sub_items = [
            sample_parameter(
                v, parent_key=f'{parent_key}.{k}', seed=seed, samples=samples
            )
            for k, v in parameter['params'].items()
        ]
        return_items.extend([sub_item for item in sub_items for sub_item in item])

    else:
        raise ConfigError(f'Parameter type {param_type} not implemented.')

    if param_type != 'parameter_collection':
        extra_keys = set(parameter.keys()).difference(set(allowed_keys))
        if len(extra_keys) > 0:
            raise ConfigError(
                f"Unexpected keys in parameter definition. Allowed keys for type '{param_type}' are "
                f'{allowed_keys}. Unexpected keys: {extra_keys}'
            )
    return return_items


def generate_grid(parameter, parent_key=''):
    """
    Generate a grid of parameter values from the input configuration.

    Parameters
    ----------
    parameter: dict
        Defines the type of parameter. Options for parameter['type'] are
            - choice: Expects a list of options in paramter['options'], which will be returned.
            - range: Expects 'min', 'max', and 'step' keys with values in the dict that are used as
                     np.arange(min, max, step)
            - uniform: Generates the grid using np.linspace(min, max, num, endpoint=True)
            - loguniform: Uniformly samples 'num' points in log space (base 10) between 'min' and 'max'
            - parameter_collection: wrapper around a dictionary of parameters (of the types above); we call this
              function recursively on each of the sub-parameters.
    parent_key: str
        The key to prepend the parameter name with. Used for nested parameters, where we here create a flattened version
        where e.g. {'a': {'b': 11}, 'c': 3} becomes {'a.b': 11, 'c': 3}

    Returns
    -------
    return_items: tuple(str, tuple(list, str))
        Name of the parameter and tuple with list containing the grid values for this parameter and zip id.

    """
    import numpy as np

    if 'type' not in parameter:
        raise ConfigError(f'No type found in parameter {parameter}')

    param_type = parameter['type']
    allowed_keys = ['type', 'zip_id']

    return_items = []

    if param_type == 'choice':
        values = parameter['options']
        allowed_keys.append('options')
        return_items.append((parent_key, values))

    elif param_type == 'range':
        min_val = parameter['min']
        max_val = parameter['max']
        step = parameter['step']
        allowed_keys.extend(['min', 'max', 'step'])
        values = list(np.arange(min_val, max_val, step))
        return_items.append((parent_key, values))

    elif param_type == 'uniform':
        min_val = parameter['min']
        max_val = parameter['max']
        num = int(parameter['num'])
        allowed_keys.extend(['min', 'max', 'num'])
        values = list(np.linspace(min_val, max_val, num, endpoint=True))
        return_items.append((parent_key, values))

    elif param_type == 'loguniform':
        min_val = parameter['min']
        max_val = parameter['max']
        num = int(parameter['num'])
        allowed_keys.extend(['min', 'max', 'num'])
        values = np.logspace(np.log10(min_val), np.log10(max_val), num, endpoint=True)
        return_items.append((parent_key, values))

    elif param_type == 'parameter_collection':
        sub_items = [
            generate_grid(v, parent_key=f'{parent_key}.{k}')
            for k, v in parameter['params'].items()
        ]
        return_items.extend([sub_item for item in sub_items for sub_item in item])

    else:
        raise ConfigError(f'Parameter {param_type} not implemented.')

    if param_type != 'parameter_collection':
        extra_keys = set(parameter.keys()).difference(set(allowed_keys))
        if len(extra_keys) > 0:
            raise ConfigError(
                f"Unexpected keys in parameter definition. Allowed keys for type '{param_type}' are "
                f'{allowed_keys}. Unexpected keys: {extra_keys}'
            )

    zip_id = parameter['zip_id'] if 'zip_id' in parameter else uuid.uuid4()
    return_items = [(item[0], (item[1], zip_id)) for item in return_items]
    return return_items


def zipped_dict(input_dict):
    """Zips dictionaries of type:
    {
        'element1': (values, zip_id),
        ...
    }
    to
    {
        'zip_id1': {
            'element1': values
            'element2': values
        },
        ...
    }

    Args:
        input_dict (dict[str, tuple(list, str)]): unzipped dictionary

    Returns:
        dict[str, dict[str, list]]: zipped dictionary
    """
    # Zip by zip_id attribute
    zipped_dict = DefaultDict(dict)
    for k, (val, zip_id) in input_dict.items():
        zipped_dict[zip_id][k] = val

    # Check that parameters in within a bundle have the same number of configurations.
    for k, bundle in zipped_dict.items():
        if len({len(x) for x in bundle.values()}) != 1:
            raise ConfigError(
                f"Parameters with zip_id '{k}' have different number of configurations!"
            )
    return zipped_dict


def cartesian_product_zipped_dict(zipped_dict):
    """Compute the Cartesian product of the ziped input dictionary values.
    Parameters
    ----------
    zipped_dict: dict of dicts of lists

    Returns
    -------
    list of dicts
        Cartesian product of the lists in the input dictionary.

    """
    zip_lengths = {
        k: len(next(iter(bundle.values()))) for k, bundle in zipped_dict.items()
    }

    for idx in itertools.product(*[range(k) for k in zip_lengths.values()]):
        yield {
            key: values[i]
            for zip_id, i in zip(zipped_dict, idx)
            for key, values in zipped_dict[zip_id].items()
        }



import copy
import itertools
import logging
import re
import subprocess
from typing import TYPE_CHECKING, Dict, List, Optional, Set

from seml.database import (
    build_filter_dict,
    delete_files,
    get_collection,
    get_database,
    get_mongodb_config,
)
from seml.experiment.config import (
    check_config,
    config_get_exclude_keys,
    generate_named_configs,
    resolve_configs,
    resolve_interpolations,
)
from seml.experiment.sources import delete_orphaned_sources, upload_sources
from seml.settings import SETTINGS
from seml.utils import (
    chunker,
    make_hash,
    s_if,
    utcnow,
)
from seml.utils.errors import MongoDBError
from seml.utils.slurm import (
    are_slurm_jobs_running,
    cancel_slurm_jobs,
    get_cluster_name,
    get_slurm_arrays_tasks,
    get_slurm_jobs,
    wait_until_slurm_jobs_finished,
)

States = SETTINGS.STATES

if TYPE_CHECKING:
    from pymongo.collection import Collection


def should_check_killed(filter_states: Optional[List[str]]) -> bool:
    """Checks whether killed experiments should be checked

    Parameters
    ----------
    filter_states : Optional[List[str]]
        The states to filter on

    Returns
    -------
    bool
        Whether killed experiments should be checked
    """
    return (
        filter_states is not None
        and len({*States.PENDING, *States.RUNNING, *States.KILLED} & set(filter_states))
        > 0
    )


def cancel_empty_pending_jobs(db_collection_name: str, *sacred_ids: int):
    """Cancels pending jobs that are not associated with any experiment

    Parameters
    ----------
    db_collection_name : str
        The collection to check for pending jobs
    sacred_ids : int
        The IDs of the experiments to check
    """
    if len(sacred_ids) == 0:
        raise ValueError('At least one sacred ID must be provided.')
    collection = get_collection(db_collection_name)
    num_pending = collection.count_documents(
        {'status': {'$in': [*States.PENDING]}, '_id': {'$in': sacred_ids}}
    )
    if num_pending > 0:
        # There are still pending experiments, we don't want to cancel the jobs.
        return
    pending_exps = list(collection.find({'_id': {'$in': sacred_ids}}, {'slurm'}))
    array_ids = {
        conf['array_id']
        for exp in pending_exps
        for conf in exp['slurm']
        if 'array_id' in conf
    }
    # Only cancel the pending jobs
    cancel_slurm_jobs(*array_ids, state='PENDING')


def cancel_jobs_without_experiments(*slurm_array_ids: str):
    """
    Cancels Slurm jobs that are not associated with any experiment that is still pending/running.

    Parameters
    ----------
    slurm_array_ids : str
        The array IDs of the Slurm jobs to check.
    """
    if len(slurm_array_ids) == 0:
        return []

    canceled_ids = []
    for array_id in slurm_array_ids:
        try:
            job_info = get_slurm_jobs(str(array_id))[0]
        except subprocess.CalledProcessError:
            # Job is not running, so we can skip this.
            continue
        col_name = job_info.get('Comment', None)
        if col_name is None:
            continue
        collection = get_collection(col_name)
        is_needed = (
            collection.count_documents(
                {
                    'slurm': {'$elemMatch': {'array_id': array_id}},
                    'status': {'$in': [*States.RUNNING, *States.PENDING]},
                },
                limit=1,
            )
            > 0
        )
        if not is_needed:
            cancel_slurm_jobs(str(array_id))
            canceled_ids.append(array_id)
    return canceled_ids


def cancel_experiment_by_id(
    collection: 'Collection',
    exp_id: int,
    set_interrupted: bool = True,
    slurm_dict: Optional[Dict] = None,
    wait: bool = False,
    timeout: int = SETTINGS.CANCEL_TIMEOUT,
):
    """Cancels a single experiment by its id

    Parameters
    ----------
    collection : str
        The collection this experiment belongs to
    exp_id : int
        The experiment id
    set_interrupted : bool, optional
        Whether to set the state of the experiment to INTERRUPTED, by default True
    slurm_dict : Optional[Dict], optional
        Optional updates to the slurm dict of the experiments, by default None
    wait : bool, optional
        Whether to wait for the cancellation by checking the slurm queue, by default False
    timeout : int, optional
        The timeout in seconds to wait for the cancellation, by default SETTINGS.CANCEL_TIMEOUT
    """

    exp = collection.find_one({'_id': exp_id})
    if exp is None:
        logging.error(f'No experiment found with ID {exp_id}.')
        return

    if slurm_dict:
        for s_conf in exp['slurm']:
            s_conf.update(slurm_dict)

    # check if the job has been scheduled at all
    array_ids = [conf.get('array_id', None) for conf in exp['slurm']]
    if any(array_id is None for array_id in array_ids):
        logging.error(f'Experiment with ID {exp_id} has not been started using Slurm.')
        return

    # check if the job has been claimed and associated with a concrete job
    is_running = 'array_id' in exp['execution']
    if is_running:
        job_str = f"{exp['execution']['array_id']}_{exp['execution']['task_id']}"
        job_strings = [job_str]
        filter_dict = {
            'execution.array_id': exp['execution']['array_id'],
            'execution.task_id': exp['execution']['task_id'],
        }
    else:
        job_strings = list(map(str, array_ids))

    # Check if slurm job exists
    if not are_slurm_jobs_running(*job_strings):
        logging.error(
            f'Slurm job {job_strings} of experiment '
            f'with ID {exp_id} is not pending/running in Slurm.'
        )
        return

    cancel_update = {
        '$set': {
            'status': States.INTERRUPTED[0],
            'stop_time': utcnow(),
        }
    }
    if set_interrupted:
        # Set the database state to INTERRUPTED
        collection.update_one({'_id': exp_id}, cancel_update)

    if is_running:
        # Check if other experiments are running in the same job
        other_exps_filter = filter_dict.copy()
        other_exps_filter['_id'] = {'$ne': exp_id}
        other_exps_filter['status'] = {'$in': [*States.RUNNING, *States.PENDING]}
        other_exp_running = collection.count_documents(other_exps_filter) >= 1

        # Cancel if no other experiments are running in the same job
        if not other_exp_running:
            cancel_slurm_jobs(job_str)
            # Wait until the job is actually gone
            if wait and not wait_until_slurm_jobs_finished(job_str, timeout=timeout):
                logging.error('Job did not cancel in time.')
                exit(1)

            if set_interrupted:
                # set state to interrupted again (might have been overwritten by Sacred in the meantime).
                collection.update_many(filter_dict, cancel_update)

    # Cancel jobs that will not execute anything
    cancel_jobs_without_experiments(*array_ids)


def cancel_experiments(
    db_collection_name: str,
    sacred_id: Optional[int] = None,
    filter_states: Optional[List[str]] = None,
    batch_id: Optional[int] = None,
    filter_dict: Optional[Dict] = None,
    yes: bool = False,
    wait: bool = False,
    confirm_threshold: int = SETTINGS.CONFIRM_THRESHOLD.CANCEL,
    timeout: int = SETTINGS.CANCEL_TIMEOUT,
):
    """Cancels experiment(s)

    Parameters
    ----------
    db_collection_name : str
        The collection to cancel experiments of
    sacred_id : Optional[int], optional
        ID of the experiment to delete. Overrides other filters.
    filter_states : Optional[List[str]], optional
        Filter on experiment states, by default None
    batch_id : Optional[int], optional
        Filter on experiment batch ids, by default None
    filter_dict : Optional[Dict], optional
        Additional filters on experiments, by default None
    yes : bool, optional
        Whether to override confirmation prompts, by default False
    wait : bool, optional
        Whether to wait for all experiments be cancelled (by checking the slurm queue), by default False
    confirm_threshold : int, optional
        The threshold for the number of experiments to cancel before asking for confirmation, by default SETTINGS.CONFIRM_THRESHOLD.CANCEL
    timeout : int, optional
        The timeout in seconds to wait for the cancellation, by default SETTINGS.CANCEL_TIMEOUT
    """
    from seml.console import prompt

    collection = get_collection(db_collection_name)
    # We check whether there are slurm jobs for which after this action no
    # RUNNING experiment remains. These slurm jobs can be killed altogether.
    # However, it is NOT possible right now to cancel a single experiment in a Slurm job with multiple
    # running experiments.
    try:
        if should_check_killed(filter_states):
            detect_killed(db_collection_name, print_detected=False)

        db_filter_dict = build_filter_dict(
            filter_states, batch_id, filter_dict, sacred_id=sacred_id
        )

        to_cancel_arr_ids = list(collection.find(db_filter_dict, {'slurm': 1}))
        ncancel = len(to_cancel_arr_ids)
        if sacred_id is not None and ncancel == 0:
            logging.error(f'No experiment found with ID {sacred_id}.')

        logging.info(f'Cancelling {ncancel} experiment{s_if(ncancel)}.')
        if ncancel >= confirm_threshold:
            if not yes and not prompt('Are you sure? (y/n)', type=bool):
                exit(1)

        running_filter = copy.deepcopy(db_filter_dict)
        running_filter = {
            **running_filter,
            'execution.cluster': get_cluster_name(),
            'execution.array_id': {'$exists': True},
        }
        running_exps = list(
            collection.find(
                running_filter,
                {'_id': 1, 'status': 1, 'execution': 1},
            )
        )
        # update database status and write the stop_time
        # this should be done early so that the experiments are not picked up by other processes
        cancel_update = {
            '$set': {
                'status': States.INTERRUPTED[0],
                'stop_time': utcnow(),
            }
        }
        collection.update_many(db_filter_dict, cancel_update)

        # Cancel pending jobs that will not execute anything
        array_ids = {
            conf['array_id']
            for exp in to_cancel_arr_ids
            for conf in exp['slurm']
            if 'array_id' in conf
        }
        canceled = cancel_jobs_without_experiments(*array_ids)

        # set of slurm IDs in the database
        slurm_ids = {
            (e['execution']['array_id'], e['execution']['task_id'])
            for e in running_exps
        }
        # set of experiment IDs to be cancelled.
        exp_ids = {e['_id'] for e in running_exps}
        to_cancel = set()

        # iterate over slurm IDs to check which slurm jobs can be cancelled altogether
        for a_id, t_id in slurm_ids:
            # find experiments RUNNING under the slurm job
            jobs_running = [
                e
                for e in running_exps
                if (
                    e['execution']['array_id'] == a_id
                    and e['execution']['task_id'] == t_id
                    and e['status'] in States.RUNNING
                )
            ]
            running_exp_ids = {e['_id'] for e in jobs_running}
            if len(running_exp_ids.difference(exp_ids)) == 0:
                # there are no running jobs in this slurm job that should not be canceled.
                to_cancel.add(f'{a_id}_{t_id}')

        # cancel all Slurm jobs for which no running experiment remains.
        if len(to_cancel) > 0:
            chunk_size = 100
            chunks = list(chunker(list(to_cancel), chunk_size))
            [cancel_slurm_jobs(*chunk) for chunk in chunks]
            # Wait until all jobs are actually stopped.
            for chunk in chunks:
                if wait and not wait_until_slurm_jobs_finished(*chunk, timeout=timeout):
                    logging.error('Job did not cancel in time.')
                    exit(1)

        canceled = list(map(str, canceled + list(to_cancel)))
        n_canceled = len(canceled)
        if n_canceled > 0:
            logging.info(
                f'Canceled job{s_if(n_canceled)} with the following ID{s_if(n_canceled)}: '
                + ', '.join(canceled)
            )
        # Let's repeat this in case a job cleaned itself up and overwrote the status.
        collection.update_many(db_filter_dict, cancel_update)
    except subprocess.CalledProcessError:
        logging.warning(
            'One or multiple Slurm jobs were no longer running when I tried to cancel them.'
        )


def delete_experiments(
    db_collection_name: str,
    sacred_id: Optional[int] = None,
    filter_states: Optional[List[str]] = None,
    batch_id: Optional[int] = None,
    filter_dict: Optional[Dict] = None,
    yes: bool = False,
    cancel: bool = True,
):
    """Deletes experiment(s).

    Parameters
    ----------
    db_collection_name : str
        The collection name to which to delete experiments from
    sacred_id : Optional[int], optional
        ID of the experiment to delete. Overrides other filters.
    filter_states : Optional[List[str]], optional
        Filter on experiment states, by default None
    batch_id : Optional[int], optional
        Filter on experiment batch ids, by default None
    filter_dict : Optional[Dict], optional
        Additional filters on experiments, by default None
    yes : bool, optional
        Whether to override confirmation prompts, by default False
    """
    from seml.console import prompt

    # Before deleting, we should first cancel the experiments that are still running.
    if cancel:
        cancel_states = set(States.PENDING + States.RUNNING)
        if filter_states is not None and len(filter_states) > 0:
            cancel_states = cancel_states.intersection(filter_states)

        if len(cancel_states) > 0:
            cancel_experiments(
                db_collection_name,
                sacred_id,
                list(cancel_states),
                batch_id,
                filter_dict,
                yes=False,
                confirm_threshold=1,
                wait=True,
            )

    collection = get_collection(db_collection_name)
    experiment_files_to_delete = []

    filter_dict = build_filter_dict(
        filter_states, batch_id, filter_dict, sacred_id=sacred_id
    )
    ndelete = collection.count_documents(filter_dict)
    if sacred_id is not None and ndelete == 0:
        raise MongoDBError(f'No experiment found with ID {sacred_id}.')
    batch_ids = collection.find(filter_dict, {'batch_id'})
    batch_ids_in_del = {x.get('batch_id', -1) for x in batch_ids}

    logging.info(
        f'Deleting {ndelete} configuration{s_if(ndelete)} from database collection.'
    )
    if ndelete >= SETTINGS.CONFIRM_THRESHOLD.DELETE:
        if not yes and not prompt('Are you sure? (y/n)', type=bool):
            exit(1)

    # Collect sources uploaded by sacred.
    exp_sources_list = collection.find(
        filter_dict, {'experiment.sources': 1, 'artifacts': 1}
    )
    for exp in exp_sources_list:
        experiment_files_to_delete.extend(get_experiment_files(exp))
    result = collection.delete_many(filter_dict)
    if not result.deleted_count == ndelete:
        logging.error(
            f'Only {result.deleted_count} of {ndelete} experiments were deleted.'
        )

    # Delete sources uploaded by sacred.
    delete_files(collection.database, experiment_files_to_delete)
    logging.info(
        f'Deleted {len(experiment_files_to_delete)} files associated with deleted experiments.'
    )

    if len(batch_ids_in_del) > 0:
        # clean up the uploaded sources if no experiments of a batch remain
        delete_orphaned_sources(collection, batch_ids_in_del)

    if collection.count_documents({}) == 0:
        collection.drop()


def drop_collections(
    pattern: str, mongodb_config: Optional[Dict] = None, yes: bool = False
):
    """
    Drops collections matching the given pattern.

    Parameters
    ----------
    pattern : str
        The regex collection names have to match against
    mongodb_config : dict or None
        A configuration for the mongodb. If None, the standard config is used.
    yes : bool
        Whether to override confirmation prompts
    """
    from seml.console import list_items, prompt

    # Get the database
    if mongodb_config is None:
        mongodb_config = get_mongodb_config()
    db = get_database(**mongodb_config)
    expression = re.compile(pattern)
    collection_names = [
        name
        for name in db.list_collection_names()
        if name not in ('fs.chunks', 'fs.files') and expression.match(name)
    ]
    if len(collection_names) == 0:
        logging.info('No collections found.')
        return
    if not yes:
        logging.info(
            f'The following {len(collection_names)} collection will be deleted:'
        )
        list_items(collection_names)
        if not prompt('Are you sure? (y/n)', type=bool):
            return
    for name in collection_names:
        delete_experiments(name, yes=True)


def reset_slurm_dict(exp: Dict):
    """Resets the slurm dict of an experiment

    Parameters
    ----------
    exp : Dict
        The experiment of which to reset the slurm dict
    """
    keep_slurm = set()
    keep_slurm.update(SETTINGS.VALID_SLURM_CONFIG_VALUES)
    for sub_conf in exp['slurm']:
        slurm_keys = set(sub_conf.keys())
        for key in slurm_keys - keep_slurm:
            del sub_conf[key]

        # Clean up sbatch_options dictionary
        remove_sbatch = {'job-name', 'output', 'array', 'comment'}
        sbatch_keys = set(sub_conf['sbatch_options'].keys())
        for key in remove_sbatch & sbatch_keys:
            del sub_conf['sbatch_options'][key]


def reset_single_experiment(collection: 'Collection', exp: Dict):
    """Resets a single experiment

    Parameters
    ----------
    collection : pymongo.collection.Collection
        The collection to which the experiment belongs to
    exp : Dict
        The experiment dict
    """
    exp['status'] = States.STAGED[0]
    # queue_time for backward compatibility.
    keep_entries = [
        'batch_id',
        'status',
        'seml',
        'slurm',
        'config',
        'config_hash',
        'add_time',
        'queue_time',
        'git',
        'config_unresolved',
    ]

    # Clean up SEML dictionary
    keep_seml = {'source_files', 'working_dir', SETTINGS.SEML_CONFIG_VALUE_VERSION}
    keep_seml.update(SETTINGS.VALID_SEML_CONFIG_VALUES)
    seml_keys = set(exp['seml'].keys())
    for key in seml_keys - keep_seml:
        del exp['seml'][key]

    reset_slurm_dict(exp)

    collection.replace_one(
        {'_id': exp['_id']},
        {entry: exp[entry] for entry in keep_entries if entry in exp},
        upsert=False,
    )


def reset_experiments(
    db_collection_name: str,
    sacred_id: Optional[int] = None,
    filter_states: Optional[List[str]] = None,
    batch_id: Optional[int] = None,
    filter_dict: Optional[Dict] = None,
    yes: bool = False,
):
    """Resets experiments

    Parameters
    ----------
    db_collection_name : str
        The name of the collection to resets experiments from
    sacred_id : Optional[int], optional
        If given, the id of the experiment to reset. Overrides other filters, by default None
    filter_states : Optional[List[str]], optional
        Filter on experiment states, by default None
    batch_id : Optional[int], optional
        Filter on the batch ID of experiments, by default None
    filter_dict : Optional[Dict], optional
        Additional filters, by default None
    yes : bool, optional
        Whether to override confirmation prompts, by default False
    """
    from seml.console import prompt

    collection = get_collection(db_collection_name)
    if should_check_killed(filter_states):
        detect_killed(db_collection_name, print_detected=False)
    filter_dict = build_filter_dict(
        filter_states, batch_id, filter_dict, sacred_id=sacred_id
    )
    nreset = collection.count_documents(filter_dict)
    exps = collection.find(filter_dict)
    if sacred_id is not None and nreset == 0:
        raise MongoDBError(f'No experiment found with ID {sacred_id}.')

    logging.info(f'Resetting the state of {nreset} experiment{s_if(nreset)}.')
    if nreset >= SETTINGS.CONFIRM_THRESHOLD.RESET:
        if not yes and not prompt('Are you sure? (y/n)', type=bool):
            exit(1)
    for exp in exps:
        reset_single_experiment(collection, exp)


def detect_killed(db_collection_name: str, print_detected: bool = True):
    """Detects killed experiments by checking the slurm status

    Parameters
    ----------
    db_collection_name : str
        The collection to check killed experiments from
    print_detected : bool, optional
        Whether to print how many killed experiments have been detected, by default True
    """
    collection = get_collection(db_collection_name)
    cluster = get_cluster_name()
    exps = collection.find(
        {
            'status': {'$in': [*States.PENDING, *States.RUNNING]},
            'execution.cluster': cluster,  # only check experiments that are running on the current cluster
            # Previously we only checked for started experiments by including the following line:
            # 'host': {'$exists': True},  # only check experiments that have been started
            # Though, this does not catch the case where a user cancels pending experiments with scancel.
            # I (Nicholas) am not 100% sure about the implications of removing the check but it at least
            # resolves the issue around manually canceled jobs.
        }
    )
    running_jobs = get_slurm_arrays_tasks()
    nkilled = 0
    for exp in exps:
        # detect whether the experiment is running in slurm
        exp_running = exp['execution'].get('array_id', -1) in running_jobs and (
            any(
                exp['execution']['task_id'] in r
                for r in running_jobs[exp['execution']['array_id']][0]
            )
            or exp['execution']['task_id']
            in running_jobs[exp['execution']['array_id']][1]
        )
        # detect whether any job that could execute it is pending
        array_ids = [conf['array_id'] for conf in exp['slurm']]
        # Any of these jobs may still pull the experiment and run it
        exp_pending = any(array_id in running_jobs for array_id in array_ids)

        if not exp_running and not exp_pending:
            if 'stop_time' in exp:
                # the experiment is already over but failed to report properly
                collection.update_one(
                    {'_id': exp['_id']}, {'$set': {'status': States.INTERRUPTED[0]}}
                )
            else:
                # the experiment was externally killed
                nkilled += 1
                collection.update_one(
                    {'_id': exp['_id']}, {'$set': {'status': States.KILLED[0]}}
                )
                try:
                    with open(exp['seml']['output_file'], errors='replace') as f:
                        all_lines = f.readlines()
                    collection.update_one(
                        {'_id': exp['_id']}, {'$set': {'fail_trace': all_lines[-4:]}}
                    )
                except OSError:
                    # If the experiment is cancelled before starting (e.g. when still queued), there is not output file.
                    logging.verbose(
                        f"File {exp['seml']['output_file']} could not be read."
                    )
                except KeyError:
                    logging.verbose(
                        f"Output file not found in experiment {exp['_id']}."
                    )
    if print_detected:
        logging.info(f'Detected {nkilled} externally killed experiment{s_if(nkilled)}.')


def get_experiment_files(experiment: Dict) -> List[str]:
    """Gets the file ids of files associated with an experiment

    Parameters
    ----------
    experiment : Dict
        The experiment dict

    Returns
    -------
    List[str]
        A list of file ids
    """
    experiment_files = []
    if 'experiment' in experiment:
        if 'sources' in experiment['experiment']:
            exp_sources = experiment['experiment']['sources']
            experiment_files.extend([x[1] for x in exp_sources])
    if 'artifacts' in experiment:
        experiment_files.extend([x['file_id'] for x in experiment['artifacts']])
    return experiment_files


def reload_sources(
    db_collection_name: str,
    batch_ids: Optional[List[int]] = None,
    keep_old: bool = False,
    yes: bool = False,
):
    """Reloads the sources of experiment(s)

    Parameters
    ----------
    db_collection_name : str
        The collection to reload sources from
    batch_ids : Optional[List[int]], optional
        Filter on the batch ids, by default None
    keep_old : bool, optional
        Whether to keep old source files in the fs, by default False
    yes : bool, optional
        Whether to override confirmation prompts, by default False
    resolve : bool, optional
        Whether to re-resolve the config values
    """
    from importlib.metadata import version

    import gridfs
    from pymongo import UpdateOne

    from seml.console import prompt

    collection = get_collection(db_collection_name)

    if batch_ids is not None and len(batch_ids) > 0:
        filter_dict = {'batch_id': {'$in': list(batch_ids)}}
    else:
        filter_dict = {}
    db_results = list(
        collection.find(
            filter_dict, {'batch_id', 'seml', 'config', 'status', 'config_unresolved'}
        )
    )
    id_to_document = {}
    for bid, documents in itertools.groupby(db_results, lambda x: x['batch_id']):
        id_to_document[bid] = list(documents)
    states = {x['status'] for x in db_results}

    if any([s in (States.RUNNING + States.PENDING + States.COMPLETED) for s in states]):
        logging.info(
            'Some of the experiments is still in RUNNING, PENDING or COMPLETED.'
        )
        if not yes and not prompt('Are you sure? (y/n)', type=bool):
            exit(1)

    for batch_id, documents in id_to_document.items():
        seml_config = documents[0]['seml']

        version_seml_config = seml_config.get(
            SETTINGS.SEML_CONFIG_VALUE_VERSION, (0, 0, 0)
        )
        version_str = '.'.join(map(str, version_seml_config))
        if version_str != version('seml'):
            logging.warn(
                f'Batch {batch_id} was added with seml version "{version_str}" '
                f'which mismatches the current version {version("seml")}'
            )

        if 'working_dir' not in seml_config or not seml_config['working_dir']:
            logging.error(f'Batch {batch_id}: No source files to refresh.')
            continue

        if any(
            document.get('config_unresolved', None) is None for document in documents
        ):
            logging.warn(
                f'Batch {batch_id}: Some experiments do not have an unresolved configuration. '
                'The resolved configuration "config" will be used for resolution instead.'
            )
        configs_unresolved = [
            document.get('config_unresolved', document['config'])
            for document in documents
        ]
        configs, named_configs = generate_named_configs(configs_unresolved)
        configs = resolve_configs(
            seml_config['executable'],
            seml_config['conda_environment'],
            configs,
            named_configs,
            seml_config['working_dir'],
        )

        # If the seed was explicited, it should be kept for the new resolved config when reloading resources
        for config, config_unresolved in zip(configs, configs_unresolved):
            if SETTINGS.CONFIG_KEY_SEED in configs_unresolved:
                config[SETTINGS.CONFIG_KEY_SEED] = config_unresolved[
                    SETTINGS.CONFIG_KEY_SEED
                ]

        documents = [
            resolve_interpolations(
                {
                    **{**document, 'config': config},
                    'config_unresolved': config_unresolved,
                }
            )
            for document, config, config_unresolved in zip(
                documents, configs, configs_unresolved
            )
        ]

        result = collection.bulk_write(
            [
                UpdateOne(
                    {'_id': document['_id']},
                    {
                        '$set': {
                            'config': document['config'],
                            'config_unresolved': document['config_unresolved'],
                            'config_hash': make_hash(
                                document['config'],
                                config_get_exclude_keys(
                                    document['config'], document['config_unresolved']
                                ),
                            ),
                        }
                    },
                )
                for document in documents
            ]
        )
        logging.info(
            f'Batch {batch_id}: Resolved configurations of {result.matched_count} experiments against new source files ({result.modified_count} changed).'
        )

        # Check whether the configurations aligns with the current source code
        check_config(
            seml_config['executable'],
            seml_config['conda_environment'],
            [document['config'] for document in documents],
            seml_config['working_dir'],
        )

        # Find the currently used source files
        db = collection.database
        fs = gridfs.GridFS(db)
        fs_filter_dict = {
            'metadata.batch_id': batch_id,
            'metadata.collection_name': f'{collection.name}',
            'metadata.deprecated': {'$exists': False},
        }
        current_source_files = db['fs.files'].find(filter_dict, '_id')
        current_ids = [x['_id'] for x in current_source_files]
        fs_filter_dict = {'_id': {'$in': current_ids}}
        # Deprecate them
        db['fs.files'].update_many(
            fs_filter_dict, {'$set': {'metadata.deprecated': True}}
        )
        try:
            # Try to upload the new ones
            source_files = upload_sources(seml_config, collection, batch_id)
        except Exception as e:
            # If it fails we reconstruct the old ones
            logging.error(
                f'Batch {batch_id}: Source import failed. Restoring old files.'
            )
            db['fs.files'].update_many(
                fs_filter_dict, {'$unset': {'metadata.deprecated': ''}}
            )
            raise e

        try:
            # Try to assign the new ones to the experiments
            filter_dict = {'batch_id': batch_id}
            collection.update_many(
                filter_dict, {'$set': {'seml.source_files': source_files}}
            )
            logging.info(f'Batch {batch_id}: Successfully reloaded source code.')
        except Exception as e:
            logging.error(f'Batch {batch_id}: Failed to set new source files.')
            # Delete new source files from DB
            for to_delete in source_files:
                fs.delete(to_delete[1])
            raise e

        # Delete the old source files
        if not keep_old:
            fs_filter_dict = {
                'metadata.batch_id': batch_id,
                'metadata.collection_name': f'{collection.name}',
                'metadata.deprecated': True,
            }
            source_files_old = [
                x['_id'] for x in db['fs.files'].find(fs_filter_dict, {'_id'})
            ]
            for to_delete in source_files_old:
                fs.delete(to_delete)


def detect_duplicates(
    db_collection_name: str, filter_dict: Optional[Dict] = None
) -> List[Set[int]]:
    """Finds duplicate configurations based on their hashes.

    Parameters
    ----------
    db_collection_name : str
        The collection to check

    Returns
    -------
    List[Set[int]]
        All duplicate experiments.
    """
    collection = get_collection(db_collection_name)
    pipeline = [
        {
            '$group': {
                '_id': '$config_hash',
                'ids': {'$addToSet': '$_id'},
                'count': {'$sum': 1},
            }
        },
        {
            '$match': {
                'count': {'$gt': 1},
            }
        },
    ]
    if filter_dict is not None:
        pipeline = [{'$match': filter_dict}] + pipeline
    duplicates = collection.aggregate(pipeline)
    return [set(duplicate['ids']) for duplicate in duplicates]



from seml.evaluation import *  # noqa
from seml.experiment import Experiment  # noqa
from seml.experiment.observers import *  # noqa
from seml.utils.module_hider import AUTOCOMPLETING


if AUTOCOMPLETING:
    __version__ = '0.0.0'
else:
    import importlib.metadata

    __version__ = importlib.metadata.version(__package__ or __name__)



import logging
import os
from typing import Dict, List, Optional

from seml.database import build_filter_dict, get_collection
from seml.experiment.sources import load_sources_from_db
from seml.settings import SETTINGS

States = SETTINGS.STATES


def download_sources(
    target_directory: str,
    collection_name: str,
    sacred_id: Optional[int] = None,
    filter_states: Optional[List[str]] = None,
    batch_id: Optional[int] = None,
    filter_dict: Optional[Dict] = None,
):
    """
    Restore source files from the database to the provided path. This is a helper function for the CLI.

    Parameters
    ----------
    target_directory: str
        The directory where the source files should be restored.
    collection_name: str
        The name of the MongoDB collection.
    sacred_id: int
        The ID of the Sacred experiment.
    filter_states: List[str]
        The states of the experiments to filter.
    batch_id: int
        The ID of the batch.
    filter_dict: Dict
        Additional filter dictionary.
    """
    from seml.console import prompt

    filter_dict = build_filter_dict(
        filter_states, batch_id, filter_dict, sacred_id=sacred_id
    )
    collection = get_collection(collection_name)
    experiments = list(collection.find(filter_dict))
    batch_ids = {exp['batch_id'] for exp in experiments}

    if len(batch_ids) > 1:
        logging.error(
            f'Multiple source code versions found for batch IDs: {batch_ids}.'
        )
        logging.error('Please specify the target experiment more concretely.')
        exit(1)

    exp = experiments[0]
    target_directory = os.path.expandvars(os.path.expanduser(target_directory))
    if not os.path.exists(target_directory):
        os.mkdir(target_directory)
    if os.listdir(target_directory):
        logging.warning(
            f'Target directory "{target_directory}" is not empty. '
            f'Files may be overwritten.'
        )
        if not prompt('Are you sure you want to continue? (y/n)', type=bool):
            exit(0)

    load_sources_from_db(exp, collection, target_directory)
    logging.info(f'Source files restored to "{target_directory}".')



import functools
import os
import subprocess
import time
from typing import Dict, Optional, Union

from seml.settings import SETTINGS


@functools.lru_cache
def get_cluster_name():
    """
    Retrieves the name of the cluster from the Slurm configuration.

    Returns
    -------
    str
        The name of the cluster
    """
    try:
        return (
            subprocess.run(
                "scontrol show config | grep ClusterName | awk '{print $3}'",
                shell=True,
                capture_output=True,
            )
            .stdout.decode()
            .strip()
        )
    except subprocess.SubprocessError:
        return 'unknown'


def get_slurm_jobs(*job_ids: str):
    """
    Returns a list of dictionaries containing information about the Slurm jobs with the given job IDs.
    If no job IDs are provided, information about all jobs is returned.

    Parameters
    ----------
    job_ids : Optional[Sequence[str]]
        The job IDs of the jobs to get information about

    Returns
    -------
    List[Dict[str, str]]
        A list of dictionaries containing information about the jobs
    """
    if len(job_ids) == 0:
        job_info_str = subprocess.run(
            'scontrol show job',
            shell=True,
            check=True,
            capture_output=True,
        ).stdout.decode('utf-8')
        job_info_strs = job_info_str.split('\n\n')
    else:
        job_info_strs = []
        for job_id in job_ids:
            job_info_str = subprocess.run(
                f'scontrol show job {job_id}',
                shell=True,
                check=True,
                capture_output=True,
            ).stdout.decode('utf-8')
            job_info_strs.append(job_info_str)
    job_info_strs = list(filter(None, job_info_strs))
    job_infos = list(map(parse_scontrol_job_info, job_info_strs))
    return job_infos


def parse_scontrol_job_info(job_info: str):
    """
    Converts the return value of `scontrol show job <jobid>` into a python dictionary.

    Parameters
    ----------
    job_info : str
        The output of `scontrol show job <jobid>`

    Returns
    -------
    dict
        The job information as a dictionary
    """
    job_info_dict: Dict[str, str] = {}
    # we may split to many times, e.g., if a value contains a space
    unfiltered_lines = job_info.split()
    filtered_lines = []
    for line in unfiltered_lines:
        if line:
            if '=' in line:
                # new variable
                filtered_lines.append(line)
            else:
                # just append to the previous variable
                filtered_lines[-1] += ' ' + line

    # Now every line must contain a '=' sign and we can simply split here
    for line in filtered_lines:
        key, value = line.split('=', 1)
        job_info_dict[key] = value
    return job_info_dict


def get_slurm_arrays_tasks(filter_by_user: bool = False):
    """Get a dictionary of running/pending Slurm job arrays (as keys) and tasks (as values)

    Parameters:
    -----------
    filter_by_user : bool
        Whether to only check jobs by the current user, by default False
    """
    try:
        squeue_cmd = f"SLURM_BITSTR_LEN=1024 squeue -a -t {','.join(SETTINGS.SLURM_STATES.ACTIVE)} -h -o %i"
        if filter_by_user:
            squeue_cmd += ' -u `whoami`'
        squeue_out = subprocess.run(
            squeue_cmd, shell=True, check=True, capture_output=True
        ).stdout
        jobs = [job_str for job_str in squeue_out.splitlines() if b'_' in job_str]
        if len(jobs) > 0:
            array_ids_str, task_ids = zip(*[job_str.split(b'_') for job_str in jobs])
            # `job_dict`: This dictionary has the job array IDs as keys and the values are
            # a list of 1) the pending job task range and 2) a list of running job task IDs.
            job_dict = {}
            for i, task_range_str in enumerate(task_ids):
                array_id = int(array_ids_str[i])
                if array_id not in job_dict:
                    job_dict[array_id] = [[range(0)], []]

                if b'[' in task_range_str:
                    # Remove brackets and maximum number of simultaneous jobs
                    task_range_str = task_range_str[1:-1].split(b'%')[0]
                    # The overall pending tasks array can be split into multiple arrays by cancelling jobs
                    job_id_ranges = task_range_str.split(b',')
                    for r in job_id_ranges:
                        if b'-' in r:
                            lower, upper = r.split(b'-')
                        else:
                            lower = upper = r
                        job_dict[array_id][0].append(range(int(lower), int(upper) + 1))
                else:
                    # Single task IDs belong to running jobs
                    task_id = int(task_range_str)
                    job_dict[array_id][1].append(task_id)

            return job_dict
        else:
            return {}
    except subprocess.CalledProcessError:
        return {}


def get_current_slurm_array_id():
    slurm_array_id = os.environ.get('SLURM_ARRAY_JOB_ID', None)
    slurm_task_id = os.environ.get('SLURM_ARRAY_TASK_ID', None)
    return slurm_array_id, slurm_task_id


def get_current_slurm_job_id():
    return os.environ.get('SLURM_JOB_ID', None)


def cancel_slurm_jobs(*job_ids: str, state: Optional[str] = None):
    """
    Cancels the Slurm jobs with the given job IDs.

    Parameters
    ----------
    job_ids : Sequence[str]
        The job IDs of the jobs to cancel
    """
    job_str = ' '.join(map(str, job_ids))
    if state is not None:
        subprocess.run(f'scancel -t {state} {job_str}', shell=True, check=False)
    else:
        subprocess.run(f'scancel {job_str}', shell=True, check=False)


def are_slurm_jobs_running(*job_ids: str):
    """
    Checks the Slurm queue to see if the jobs with the given job IDs are still running.

    Parameters
    ----------
    job_ids : Sequence[str]
        The job IDs of the jobs to check

    Returns
    -------
    bool
        True if the jobs are still running, False otherwise
    """
    return (
        len(
            subprocess.run(
                f"squeue -h -o '%A' --jobs={','.join(job_ids)}",
                shell=True,
                check=True,
                capture_output=True,
            ).stdout
        )
        > 0
    )


def wait_until_slurm_jobs_finished(*job_ids: str, timeout: Union[int, float]):
    """
    Waits until all jobs are finished or until the timeout is reached.

    Parameters
    ----------
    job_ids: Sequence[str]
        The job IDs of the jobs to wait for
    timeout: Union[int, float]
        The maximum time to wait in seconds

    Returns
    -------
    bool
        True if the jobs finished before the timeout, False otherwise
    """
    end_time = time.time() + timeout
    while are_slurm_jobs_running(*job_ids):
        time.sleep(0.1)
        if time.time() > end_time:
            return False
    return True



from seml.utils.module_hider import ModuleHider

# If we are in a completion shell, we don't want to import rich.
# due to its high import time.
with ModuleHider('rich'):
    from typer import *  # noqa



import json
import re
from datetime import datetime, timedelta, timezone

from bson import json_util
from sacred.config.config_files import load_config_file
from sacred.observers.base import RunObserver, td_format

from seml.utils.json import NumpyEncoder


def to_local_timezone(dtime):
    return dtime.replace(tzinfo=timezone.utc).astimezone(tz=None)


class MattermostObserver(RunObserver):
    """
    Based on Sacred's Slack observer: https://github.com/IDSIA/sacred/blob/master/sacred/observers/slack.py
    Sends a message to Mattermost upon completion/failing of an experiment.
    """

    @classmethod
    def from_config(cls, filename):
        """
        Create a MattermostObserver from a given configuration file.

        The file can be in any format supported by Sacred
        (.json, .pickle, [.yaml]).
        It has to specify a ``webhook_url`` and can optionally set
        ``bot_name``, ``icon``, ``completed_text``, ``interrupted_text``, and
        ``failed_text``.
        """
        return cls(**load_config_file(filename))

    def __init__(
        self,
        webhook_url,
        channel=None,
        bot_name='sacredbot',
        icon=':angel:',
        notify_on_started=False,
        notify_on_completed=True,
        notify_on_failed=True,
        notify_on_interrupted=False,
        heartbeat_interval=None,
        started_text=None,
        completed_text=None,
        failed_text=None,
        interrupted_text=None,
        heartbeat_text=None,
        convert_utc_to_local_timezone=True,
    ):
        """
        Create a Sacred observer that will send notifications to Mattermost.
        Parameters
        ----------
        webhook_url: str
            The webhook for the bot account.
        channel: str
            The channel to which to send notifications. To send direct messages, set to @username.
        bot_name: str
            The name of the bot.
        icon: str
            The icon of the bot.
        notify_on_started: bool
            Whether to send a notification when the experiment starts.
        notify_on_completed: bool
            Whether to send a notification upon completion.
        notify_on_failed: bool
            Whether to send a notification when the experiment fails.
        notify_on_interrupted: bool
            Whether to send a notification when the experiment is interrupted.
        heartbeat_interval: str
            String in the format D-hh:mm indicating how often to send heartbeat notifications. If None, send no
            notifications.
        started_text: str
            Text to be sent when the experiment starts. If None, this default will be used:
            ":hourglass_flowing_sand: *{experiment[name]}* "
            "started on host `{host_info[hostname]}` at _{start_time}_."
        completed_text: str
            Text to be sent upon completion. If None, this default will be used:
            ":white_check_mark: *{experiment[name]}* "
            "completed after _{elapsed_time}_ with result: \n```json\n{result}\n````\n"
        failed_text: str
            Text to be sent upon failure. If None, this default will be used:
            ":x: *{experiment[name]}* failed after " "_{elapsed_time}_ with `{error}`"
        interrupted_text: str
            Text to be sent upon interruption. If None, this default will be used:
            ":warning: *{experiment[name]}* " "interrupted after _{elapsed_time}_"
        heartbeat_text: str
            Text to be sent to notify that the experiment is still running. If None, this default will be used:
            ":heartpulse: *{experiment[name]}* has been up and running for " "_{elapsed_time}_. "
            "Current info dict: \n```json\n{info}\n```\n"
            "Next heartbeat will be sent in about _{heartbeat_interval}_, i.e., on _{next_heartbeat_date}_."
        convert_utc_to_local_timezone: bool
            Whether to convert UTC times to local timezone in the notifications.
        """

        self.webhook_url = webhook_url
        self.bot_name = bot_name
        self.icon = icon
        self.completed_text = completed_text or (
            ':white_check_mark: *{experiment[name]}* '
            'completed after _{elapsed_time}_ with result: \n```json\n{result}\n````\n'
        )
        self.started_text = started_text or (
            ':hourglass_flowing_sand: *{experiment[name]}* '
            'started on host `{host_info[hostname]}` at _{start_time}_.'
        )
        self.interrupted_text = interrupted_text or (
            ':warning: *{experiment[name]}* ' 'interrupted after _{elapsed_time}_'
        )
        self.failed_text = failed_text or (
            ':x: *{experiment[name]}* failed after ' '_{elapsed_time}_ with `{error}`'
        )
        self.heartbeat_text = heartbeat_text or (
            ':heartpulse: *{experiment[name]}* has been up and running for '
            '_{elapsed_time}_. '
            'Current info dict: \n```json\n{info}\n```\n'
            'Next heartbeat will be sent in about _{heartbeat_interval}_, i.e., on _{next_heartbeat_date}_.'
        )

        self.run = None
        self.channel = channel

        self.notify_on_completed = notify_on_completed
        self.notify_on_failed = notify_on_failed
        self.notify_on_interrupted = notify_on_interrupted
        self.notify_on_started = notify_on_started
        self.notify_on_heartbeat = False
        self.last_heartbeat_notification = None
        self.convert_utc_to_local_timezone = convert_utc_to_local_timezone

        self.heartbeat_interval = None
        if heartbeat_interval is not None:
            # unfortunately datetime.strptime() doesn't work with timedeltas, so we parse the date ourselves:
            pattern = re.compile('([0-9]+)-([0-9]+):([0-9]+)')
            days, hours, minutes = pattern.match(heartbeat_interval).groups()
            self.heartbeat_interval = timedelta(
                days=int(days), hours=int(hours), minutes=int(minutes)
            )
            self.notify_on_heartbeat = True

    def started_event(
        self, ex_info, command, host_info, start_time, config, meta_info, _id
    ):
        import requests

        if self.convert_utc_to_local_timezone:
            start_time = to_local_timezone(start_time)

        self.run = {
            '_id': _id,
            'config': config,
            'start_time': start_time,
            'experiment': ex_info,
            'command': command,
            'host_info': host_info,
        }
        if self.heartbeat_interval is not None:
            self.run['heartbeat_interval'] = td_format(self.heartbeat_interval)
            self.last_heartbeat_notification = start_time

        if not self.notify_on_started:
            return

        data = {
            'username': self.bot_name,
            'icon_emoji': self.icon,
            'text': self.get_started_text(),
        }

        if self.channel is not None:
            data['channel'] = self.channel
        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
        requests.post(self.webhook_url, data=json.dumps(data), headers=headers)

    def get_completed_text(self):
        return self.completed_text.format(**self.run)

    def get_started_text(self):
        return self.started_text.format(**self.run)

    def get_interrupted_text(self):
        return self.interrupted_text.format(**self.run)

    def get_failed_text(self):
        return self.failed_text.format(**self.run)

    def get_heartbeat_text(self):
        return self.heartbeat_text.format(**self.run)

    def completed_event(self, stop_time, result):
        import requests

        if self.completed_text is None or not self.notify_on_completed:
            return

        if self.convert_utc_to_local_timezone:
            stop_time = to_local_timezone(stop_time)

        self.run['result'] = json.dumps(result, indent=4, cls=NumpyEncoder)
        self.run['stop_time'] = stop_time
        self.run['elapsed_time'] = td_format(stop_time - self.run['start_time'])

        data = {
            'username': self.bot_name,
            'icon_emoji': self.icon,
            'text': self.get_completed_text(),
        }
        if self.channel is not None:
            data['channel'] = self.channel
        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
        requests.post(self.webhook_url, data=json.dumps(data), headers=headers)

    def interrupted_event(self, interrupt_time, status):
        import requests

        if self.interrupted_text is None or not self.notify_on_interrupted:
            return

        if self.convert_utc_to_local_timezone:
            interrupt_time = to_local_timezone(interrupt_time)

        self.run['status'] = status
        self.run['interrupt_time'] = interrupt_time
        self.run['elapsed_time'] = td_format(interrupt_time - self.run['start_time'])

        data = {
            'username': self.bot_name,
            'icon_emoji': self.icon,
            'text': self.get_interrupted_text(),
        }
        if self.channel is not None:
            data['channel'] = self.channel
        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
        requests.post(self.webhook_url, data=json.dumps(data), headers=headers)

    def failed_event(self, fail_time, fail_trace):
        import requests

        if self.failed_text is None or not self.notify_on_failed:
            return
        if self.convert_utc_to_local_timezone:
            fail_time = to_local_timezone(fail_time)

        self.run['fail_trace'] = '\n'.join(fail_trace)
        self.run['error'] = fail_trace[-1].strip()
        self.run['fail_time'] = fail_time
        self.run['elapsed_time'] = td_format(fail_time - self.run['start_time'])

        data = {
            'username': self.bot_name,
            'icon_emoji': self.icon,
            'text': self.get_failed_text(),
        }
        if self.channel is not None:
            data['channel'] = self.channel
        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
        requests.post(self.webhook_url, data=json.dumps(data), headers=headers)

    def heartbeat_event(self, info, captured_out, beat_time, result):
        import requests

        if self.heartbeat_text is None or not self.notify_on_heartbeat:
            return

        if self.convert_utc_to_local_timezone:
            beat_time = to_local_timezone(beat_time)

        if beat_time < self.last_heartbeat_notification + self.heartbeat_interval:
            return

        next_heartbeat_notification = beat_time + self.heartbeat_interval
        self.run['next_heartbeat_date'] = datetime.strftime(
            next_heartbeat_notification, '%Y-%m-%d %H:%M'
        )
        self.run['elapsed_time'] = td_format(beat_time - self.run['start_time'])
        self.run['info'] = json.dumps(info, indent=4, default=json_util.default)

        data = {
            'username': self.bot_name,
            'icon_emoji': self.icon,
            'text': self.get_heartbeat_text(),
        }
        if self.channel is not None:
            data['channel'] = self.channel
        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
        requests.post(self.webhook_url, data=json.dumps(data), headers=headers)
        self.last_heartbeat_notification = beat_time



import copy
import functools
import hashlib
import logging
import os
import subprocess
import time
from contextlib import contextmanager
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)


def s_if(n: int) -> str:
    return '' if n == 1 else 's'


def unflatten(dictionary: dict, sep: str = '.', recursive: bool = False, levels=None):
    """
    Turns a flattened dict into a nested one, e.g. {'a.b':2, 'c':3} becomes {'a':{'b': 2}, 'c': 3}
    From https://stackoverflow.com/questions/6037503/python-unflatten-dict.

    Parameters
    ----------
    dictionary: dict to be un-flattened
    sep: separator with which the nested keys are separated
    recursive: bool, default: False
        Whether to also un-flatten sub-dictionaries recursively. NOTE: if recursive is True, there can be key
        collisions, e.g.: {'a.b': 3, 'a': {'b': 5}}. In these cases, keys which are later in the insertion order
        overwrite former ones, i.e. the example above returns {'a': {'b': 5}}.
    levels: int or list of ints (optional).
        If specified, only un-flatten the desired levels. E.g., if levels= [0, -1], then {'a.b.c.d': 111} becomes
        {'a': {'b.c': {'d': 111}}}.

    Returns
    -------
    result_dict: the nested dictionary.
    """

    duplicate_key_warning_str = (
        'Duplicate key detected in recursive dictionary unflattening. '
        "Overwriting previous entries of '{}'."
    )

    if levels is not None:
        if not isinstance(levels, tuple) and not isinstance(levels, list):
            levels = [levels]
        if len(levels) == 0:
            raise ValueError(
                'Need at least one level to unflatten when levels != None.'
            )
        if not isinstance(levels[0], int):
            raise TypeError(
                f'Levels must be list or set of integers, got type {type(levels[0])}.'
            )

    result_dict = dict()
    for key, value in dictionary.items():
        if isinstance(value, dict) and recursive:
            value = unflatten(value, sep=sep, recursive=True, levels=levels)

        parts = key.split(sep)
        if levels is not None:
            key_levels = levels.copy()
            for ix in range(len(key_levels)):
                if key_levels[ix] < 0:
                    new_ix = len(parts) + key_levels[ix] - 1
                    if (
                        key_levels[ix] == -1
                    ):  # special case so that indexing with -1 never throws an error.
                        new_ix = max(0, new_ix)
                    if new_ix < 0:
                        raise IndexError(
                            f'Dictionary key level out of bounds. ({new_ix} < 0).'
                        )
                    key_levels[ix] = new_ix
                if key_levels[ix] >= len(parts):
                    raise IndexError(
                        f'Dictionary key level {key_levels[ix]} out of bounds for size {len(parts)}.'
                    )
            key_levels = sorted(key_levels)

            key_levels = list(set(key_levels))
            new_parts = []
            ix_current = 0
            for level in key_levels:
                new_parts.append(sep.join(parts[ix_current : level + 1]))
                ix_current = level + 1

            if ix_current < len(parts):
                new_parts.append(sep.join(parts[ix_current::]))
            parts = new_parts

        d = result_dict
        # Index the existing dictionary in a nested way via the separated key levels. Create empty dicts if necessary.
        for part in parts[:-1]:
            if part not in d:
                d[part] = dict()
            elif not isinstance(d[part], dict):
                # Here we have a case such as: {'a.b': ['not_dict'], 'a': {'b': {'c': 111}}}
                # Since later keys overwrite former ones, we replace the value for {'a.b'} with {'c': 111}.
                logging.warning(duplicate_key_warning_str.format(part))
                d[part] = dict()
            # Select the sub-dictionary for the key level.
            d = d[part]
        last_key = parts[-1]
        if last_key in d:
            if isinstance(value, dict):
                intersection = set(d[last_key].keys()).intersection(value.keys())
                if len(intersection) > 0:
                    logging.warning(duplicate_key_warning_str.format(last_key))
                # Merge dictionaries, overwriting any existing values for duplicate keys.
                d[last_key] = merge_dicts(d[last_key], value)
            else:
                logging.warning(duplicate_key_warning_str.format(last_key))
                d[last_key] = value
        else:
            d[last_key] = value
    return result_dict


def flatten(dictionary: dict, parent_key: str = '', sep: str = '.'):
    """
    Flatten a nested dictionary, e.g. {'a':{'b': 2}, 'c': 3} becomes {'a.b':2, 'c':3}.
    From https://stackoverflow.com/questions/6027558/flatten-nested-dictionaries-compressing-keys

    Parameters
    ----------
    dictionary: dict to be flattened
    parent_key: string to prepend the key with
    sep: level separator

    Returns
    -------
    flattened dictionary.
    """
    import collections

    items = []
    for k, v in dictionary.items():
        k = str(k)
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.abc.MutableMapping):
            # This covers the edge case that someone supplies an empty dictionary as parameter
            if len(v) == 0:
                items.append((new_key, v))
            else:
                items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def get_from_nested(d: Dict, key: str, sep: str = '.') -> Any:
    """Gets a value from an unflattened dict, e.g. allows to use strings like `config.data` on a nesteddict

    Parameters
    ----------
    d : Dict
        The dict from which to get
    key : str
        A path to the value, separated by `sep`
    sep : str, optional
        The separator for levels in the nested dict, by default '.'

    Returns
    -------
    Any
        The nested value
    """
    for k in key.split(sep):
        d = d[k]
    return d


def list_is_prefix(first: List, second: List) -> bool:
    return len(first) <= len(second) and all(x1 == x2 for x1, x2 in zip(first, second))


def resolve_projection_path_conflicts(
    projection: Dict[str, Union[bool, int]], sep: str = '.'
) -> Dict[str, bool]:
    """Removes path conflicts in a MongoDB projection dict. E.g. if you pass the dict
    `{'config' : 1, 'config.dataset' : 1}`, MongoDB will throw an error. This method will ensure that
    always the "bigger" projection is returned, i.e. `"config"` in the aforementioned example.
    Note that this resolution will not work if you pass e.g. `{'config' : 1, 'config.dataset' : 0}`.

    Parameters
    ----------
    projection : Dict[str, bool]
        The projection to resolve
    sep : str, optional
        The separator for nested config values, by default '.'

    Returns
    -------
    Dict[str, bool]
        The resolved projection
    """
    result = {}
    for k, v in projection.items():
        k = tuple(k.split(sep))
        add_k = True
        for other in list(result.keys()):
            if list_is_prefix(k, other):
                # If `k` is a prefix of any path in `result`, this path will be removed
                if result[other] != v:
                    raise ValueError(
                        f'Can not resolve projection {(k, v), (other, result[other])}'
                    )
                del result[other]
            elif list_is_prefix(other, k):
                # If any other path in `result` is a prefix of `k` we do not add k
                if result[other] != v:
                    raise ValueError(
                        f'Can not resolve projection {(k, v), (other, result[other])}'
                    )
                add_k = False
        if add_k:
            result[k] = v
    return {sep.join(k): v for k, v in result.items()}


def chunker(seq, size):
    """
    Chunk a list into chunks of size `size`.
    From
    https://stackoverflow.com/questions/434287/what-is-the-most-pythonic-way-to-iterate-over-a-list-in-chunks

    Parameters
    ----------
    seq: input list
    size: size of chunks

    Returns
    -------
    The list of lists of size `size`
    """
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


def merge_dicts(dict1, dict2):
    """Recursively merge two dictionaries.

    Values in dict2 override values in dict1. If dict1 and dict2 contain a dictionary as a
    value, this will call itself recursively to merge these dictionaries.
    This does not modify the input dictionaries (creates an internal copy).

    Parameters
    ----------
    dict1: dict
        First dict.
    dict2: dict
        Second dict. Values in dict2 will override values from dict1 in case they share the same key.

    Returns
    -------
    return_dict: dict
        Merged dictionaries.

    """
    if not isinstance(dict1, dict):
        raise ValueError(f'Expecting dict1 to be dict, found {type(dict1)}.')
    if not isinstance(dict2, dict):
        raise ValueError(f'Expecting dict2 to be dict, found {type(dict2)}.')

    return_dict = copy.deepcopy(dict1)

    for k, v in dict2.items():
        if k not in dict1:
            return_dict[k] = v
        else:
            if isinstance(v, dict) and isinstance(dict1[k], dict):
                return_dict[k] = merge_dicts(dict1[k], dict2[k])
            else:
                return_dict[k] = dict2[k]

    return return_dict


def remove_keys_from_nested(d: Dict, keys: List[str] = []) -> Dict:
    """Removes keys from a nested dictionary

    Parameters
    ----------
    d : Dict
        the dict to remove keys from
    keys : List[str], optional
        the keys to remove, by default []. Prefixes are also allowed.

    Returns
    -------
    Dict
        a copy of the dict without the values in `keys`
    """
    return unflatten(
        {
            k: v
            for k, v in flatten(d).items()
            if not any(k.startswith(key) for key in keys)
        }
    )


def make_hash(
    d: Dict,
    exclude_keys: List[str] = [
        'seed',
    ],
):
    """
    Generate a hash for the input dictionary.
    From: https://stackoverflow.com/a/22003440
    Parameters
    ----------
    d : Dict
        The dictionary to hash
    exclude_keys : List[str]
        Keys to not hash.

    Returns
    -------
    hash (hex encoded) of the input dictionary.
    """
    import hashlib
    import json

    return hashlib.md5(
        json.dumps(remove_keys_from_nested(d, exclude_keys), sort_keys=True).encode(
            'utf-8'
        )
    ).hexdigest()


def add_logging_level(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    From https://stackoverflow.com/a/35804945
    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError(f'{levelName} already defined in logging module.')
    if hasattr(logging, methodName):
        raise AttributeError(f'{methodName} already defined in logging module.')
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError(f'{methodName} already defined in logger class.')

    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


add_logging_level('VERBOSE', 19)


class LoggingFormatter(logging.Formatter):
    FORMATS = {
        logging.INFO: '%(msg)s',
        logging.VERBOSE: '%(msg)s',
        logging.DEBUG: 'DEBUG: %(module)s: %(lineno)d: %(msg)s',
        'DEFAULT': '%(levelname)s: %(msg)s',
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno, self.FORMATS['DEFAULT'])
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class Hashabledict(dict):
    def __hash__(self):
        import json

        return hash(json.dumps(self, sort_keys=True))


@contextmanager
def working_directory(path: Path):
    """
    Context manager to temporarily change the working directory.

    Parameters
    ----------
    path: Path
        Path to the new working directory.
    """
    origin = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)


R = TypeVar('R')


class DiskCachedFunction(Generic[R]):
    def __init__(
        self,
        fun: Callable[[], R],
        name: str,
        time_to_live: float,
    ):
        from seml.settings import SETTINGS

        user = os.environ.get('USER', 'unknown')
        install_hash = hashlib.md5(name.encode('utf-8')).hexdigest()
        file_name = f'seml_{user}_{name}_{install_hash}.json'
        self.cache_path = Path(SETTINGS.TMP_DIRECTORY) / file_name
        self.time_to_live = time_to_live
        self.fun = fun

    def __call__(self) -> R:
        import json
        import time

        # Load from cache
        # if it fails or is expired we will compute it again
        if self.cache_path.exists():
            try:
                with open(self.cache_path) as f:
                    cache = json.load(f)
                if cache['expire'] > time.time():
                    return cache['result']
            except OSError:
                pass
            except json.JSONDecodeError:
                pass
        # Compute and save to cache
        result = self.fun()
        cache = {'result': result, 'expire': time.time() + self.time_to_live}
        try:
            with open(self.cache_path, 'w') as f:
                json.dump(cache, f)
        except OSError:
            # If the writing fails for any reason we can just continue.
            pass
        return result

    def clear_cache(self):
        if self.cache_path.exists():
            try:
                os.remove(self.cache_path)
                return True
            except OSError:
                return False
        return False

    def recompute_cache(self):
        if self.clear_cache():
            self()
            return True
        return False


def cache_to_disk(name: str, time_to_live: float):
    """
    Cache the result of a function to disk.

    Parameters
    ----------
    name: str
        Name of the cache file.
    time_to_live: float
        Time to live of the cache in seconds.

    Returns
    -------
    A function decorator.
    """

    def wrapper(fun: Callable[[], R]):
        return DiskCachedFunction(fun, name, time_to_live)

    return wrapper


def to_slices(items: List[int]) -> List[Tuple[int, int]]:
    """
    Convert a list of integers to a list of slices.

    Parameters
    ----------
    items: List[int]
        List of integers.

    Returns
    -------
    List[Tuple[int, int]]
        List of slices.
    """
    slices = []
    if len(items) == 0:
        return slices
    items = sorted(items)
    start, end = items[0], items[0]
    for i in items[1:]:
        if i == end + 1:
            end = i
        else:
            slices.append((start, end))
            start, end = i, i
    # last slice
    slices.append((start, end))
    return slices


def slice_to_str(s: Tuple[int, int]) -> str:
    """
    Convert a slice to a string.

    Parameters
    ----------
    s: Tuple[int, int]
        The slice.

    Returns
    -------
    str
        The slice as a string.
    """
    if s[0] == s[1]:
        return str(s[0])
    else:
        return f'{s[0]}-{s[1]}'


def to_hashable(x: Any) -> Any:
    """Returns a hashable representation of an object. Currently supports dicts and other iterables (which will be
    transformed into tuples)

    Parameters
    ----------
    x : Any
        the object to transform

    Returns
    -------
    Any
        the hashable representation
    """
    if isinstance(x, Hashable):
        return x
    elif isinstance(x, Dict):
        Hashabledict((k, to_hashable(v)) for k, v in x.items())
    elif isinstance(x, Iterable):
        return tuple(map(to_hashable, x))
    else:
        raise ValueError(f'{x} of type {type(x)} is not hashable.')


T = TypeVar('T', bound=Callable)


def warn_multiple_calls(warning: str, warn_after: int = 1):
    """
    Decorator to warn if a function is called multiple times.

    Parameters
    ----------
    warning: str
        The warning message.
    warn_after: int
        The number of calls after which to warn.
    """

    def decorator(f: T) -> T:
        num_calls = 0

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            nonlocal num_calls
            num_calls += 1
            if num_calls > warn_after:
                logging.warning(warning.format(num_calls=num_calls))
            return f(*args, **kwargs)

        return wrapper

    return decorator


def load_text_resource(path: Union[str, Path]):
    """
    Read a text resource from the package.

    Parameters
    ----------
    path: str | Path
        Path to the resource.

    Returns
    -------
    str
        The resource content.
    """
    try:
        import importlib.resources

        full_path = importlib.resources.files('seml') / path
    except (AttributeError, ImportError):
        # Python 3.8
        import importlib_resources

        full_path = importlib_resources.files('seml') / path

    with open(full_path) as inp:
        return inp.read()


def assert_package_installed(package: str, error: str):
    """
    Assert that a package is installed.

    Parameters
    ----------
    package: str
        The package name.
    """
    import importlib

    try:
        importlib.import_module(package)
    except ImportError:
        logging.error(error)
        exit(1)


def src_layout_to_flat_layout(original_path: Union[Path, str]):
    """
    Removes the first "src" directory from the path, handling any position.
    For the imports to prefer our loaded seml version, we need to convert the src-layout to the flat-layout.
    https://packaging.python.org/en/latest/discussions/src-layout-vs-flat-layout/

    Parameters:
    ----------
    original_path: Path | str
        The original path.

    Returns:
        A new path object with "src" removed.
    """
    path = Path(original_path)
    if path.parts[0] == 'src':
        return Path(*path.parts[1:])
    return path


def find_jupyter_host(
    log_file: Union[str, Path], wait: bool
) -> Tuple[Optional[str], Optional[bool]]:
    """
    Extracts the hostname from the jupyter log file and returns the URL.

    Parameters
    ----------
    log_file: str | Path
        The path to the log file.
    wait: bool
        Whether to wait until the jupyter server is running.

    Returns
    -------
    Optional[str]
        The URL of the jupyter server.
    Optional[bool]
        Whether the hostname is known. If None is returned, an error occured.
    """
    hosts_str = subprocess.run(
        'sinfo -h -o "%N|%o"', shell=True, check=True, capture_output=True
    ).stdout.decode('utf-8')
    hosts = {
        h.split('|')[0]: h.split('|')[1] for h in hosts_str.split('\n') if len(h) > 1
    }
    # Wait until jupyter is running
    if wait:
        log_file_contents = ''
        while ' is running at' not in log_file_contents:
            if os.path.exists(log_file):
                with open(log_file) as f:
                    log_file_contents = f.read()
            time.sleep(0.5)
    else:
        if not os.path.exists(log_file):
            return None, None
        with open(log_file) as f:
            log_file_contents = f.read()
        if ' is running at' not in log_file_contents:
            return None, None
    # Determine hostname
    JUPYTER_LOG_HOSTNAME_PREFIX = 'SLURM assigned me the node(s): '
    hostname = (
        [x for x in log_file_contents.split('\n') if JUPYTER_LOG_HOSTNAME_PREFIX in x][
            0
        ]
        .split(':')[1]
        .strip()
    )
    if hostname in hosts:
        hostname = hosts[hostname]
        known_host = True
    else:
        known_host = False
    # Obtain general URL
    log_file_split = log_file_contents.split('\n')
    url_lines = [x for x in log_file_split if 'http' in x]
    url = url_lines[0].split(' ')
    url_str = None
    for s in url:
        if s.startswith('http://') or s.startswith('https://'):
            url_str = s
            break
    if url_str is None:
        return log_file_contents, None
    url_str = hostname + ':' + url_str.split(':')[-1]
    url_str = url_str.rstrip('/')
    if url_str.endswith('/lab'):
        url_str = url_str[:-4]
    return url_str, known_host


def is_local_file(filename, root_dir):
    """
    See https://github.com/IDSIA/sacred/blob/master/sacred/dependencies.py
    Parameters
    ----------
    filename
    root_dir

    Returns
    -------

    """
    file_path = Path(filename).expanduser().resolve()
    root_path = Path(root_dir).expanduser().resolve()
    return root_path in file_path.parents


def smaller_than_version_filter(version: Tuple[int, int, int]):
    """
    Returns a mongodb filter that selects experiments where the version number
    is small or equal to the supplied version.

    Parameters
    ----------
    version: Tuple[int, int, int]
        The version number to compare to.

    Returns
    -------
    Dict
        The filter.
    """
    from seml.settings import SETTINGS

    version_prefix = f'seml.{SETTINGS.SEML_CONFIG_VALUE_VERSION}'
    return {
        '$or': [
            {f'{version_prefix}.0': {'$lt': version[0]}},
            {
                f'{version_prefix}.0': {'$eq': version[0]},
                f'{version_prefix}.1': {'$lt': version[1]},
            },
            {
                f'{version_prefix}.0': {'$eq': version[0]},
                f'{version_prefix}.1': {'$eq': version[1]},
                f'{version_prefix}.2': {'$lt': version[2]},
            },
        ]
    }


def utcnow():
    """
    Wrapper around datetime.datetime.now(datetime.UTC) but supports older python versions.

    Returns
    -------
    datetime.datetime
        The current datetime.
    """
    import datetime

    try:
        return datetime.datetime.now(datetime.UTC)
    except AttributeError:
        return datetime.datetime.utcnow()



import importlib
import logging
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Set, Union, cast

from seml.database import delete_files, upload_file
from seml.settings import SETTINGS
from seml.utils import is_local_file, src_layout_to_flat_layout, working_directory
from seml.utils.errors import ExecutableError, MongoDBError

if TYPE_CHECKING:
    from pymongo.collection import Collection

States = SETTINGS.STATES


def import_exe(executable, conda_env, working_dir):
    """Import the given executable file.

    Parameters
    ----------
    executable: str
        The Python file containing the experiment.
    conda_env: str
        The experiment's Anaconda environment.

    Returns
    -------
    The module of the imported executable.

    """
    # Check if current environment matches experiment environment
    if conda_env is not None and conda_env != os.environ.get('CONDA_DEFAULT_ENV'):
        logging.warning(
            f"Current Anaconda environment does not match the experiment's environment ('{conda_env}')."
        )

    with working_directory(working_dir):
        # Get experiment as module (which causes Sacred not to start ex.automain)
        exe_path = str(Path(executable).expanduser().resolve())
        sys.path.insert(0, os.path.dirname(exe_path))
        orig_handlers = logging.root.handlers
        orig_loglevel = logging.root.level
        exe_module = importlib.import_module(
            os.path.splitext(os.path.basename(executable))[0]
        )

    if exe_module.__file__ != exe_path:
        logging.error(
            f'Imported module path\n"{exe_module.__file__}" does not match executable path\n'
            f'"{exe_path}".\nIs the executable file name "{os.path.basename(executable)}" '
            f'a package name required by seml, e.g., "numpy.py"? '
            f'If yes, this case it not supported; please rename your script.\n'
            f'Otherwise, you can also skip source file uploading and configuration sanity checking '
            f'py passing "--no-code-checkpoint" and "--no-sanity-check" to seml.'
        )
        exit(1)
    logging.root.handlers = orig_handlers
    logging.root.setLevel(orig_loglevel)
    del sys.path[0]

    return exe_module


def get_imported_sources(
    executable, root_dir, conda_env, working_dir, stash_all_py_files: bool
) -> Set[str]:
    """Get the sources imported by the given executable.

    Args:
        executable (_type_): The Python file containing the experiment.
        root_dir (_type_): The root directory of the experiment.
        conda_env (_type_): The experiment's Anaconda environment.
        working_dir (_type_): The working directory of the experiment.
        stash_all_py_files (_type_): Whether to stash all .py files in the working directory.

    Returns:
        List[str]: The sources imported by the given executable.
    """
    import_exe(executable, conda_env, working_dir)
    root_path = str(Path(root_dir).expanduser().resolve())

    sources = set()
    source_added = True
    while source_added:
        source_added = False
        for name, mod in list(sys.modules.items()):
            if mod is None:
                continue
            if not getattr(mod, '__file__', False):
                continue
            filename = os.path.abspath(mod.__file__)
            if filename not in sources and is_local_file(filename, root_path):
                sources.add(filename)
                source_added = True

    if stash_all_py_files:
        for file in Path(working_dir).glob('**/*.py'):
            sources.add(str(file))

    return sources


def upload_sources(seml_config, collection, batch_id):
    with working_directory(seml_config['working_dir']):
        root_dir = str(Path(seml_config['working_dir']).expanduser().resolve())

        sources = get_imported_sources(
            seml_config['executable'],
            root_dir=root_dir,
            conda_env=seml_config['conda_environment'],
            working_dir=seml_config['working_dir'],
            stash_all_py_files=seml_config.get('stash_all_py_files', False),
        )
        executable_abs = str(Path(seml_config['executable']).expanduser().resolve())

        if executable_abs not in sources:
            raise ExecutableError(
                f'Executable {executable_abs} was not found in the source code files to upload.'
            )

        uploaded_files = []
        for s in sources:
            file_id = upload_file(s, collection, batch_id, 'source_file')
            source_path = Path(s)
            uploaded_files.append((str(source_path.relative_to(root_dir)), file_id))
    return uploaded_files


def get_git_info(filename, working_dir):
    """
    Get the git commit info.
    See https://github.com/IDSIA/sacred/blob/c1c19a58332368da5f184e113252b6b0abc8e33b/sacred/dependencies.py#L400

    Parameters
    ----------
    filename: str

    Returns
    -------
    path: str
        The base path of the repository
    commit: str
        The commit hash
    is_dirty: bool
        True if there are uncommitted changes in the repository
    """

    try:
        from git import InvalidGitRepositoryError, Repo
    except ImportError:
        logging.warning(
            'Cannot import git (pip install GitPython). ' 'Not saving git status.'
        )

    with working_directory(working_dir):
        directory = os.path.dirname(filename)

        try:
            repo = Repo(directory, search_parent_directories=True)
        except InvalidGitRepositoryError:
            return None, None, None
        try:
            path = repo.remote().url
        except ValueError:
            path = 'git:/' + repo.working_dir
        commit = repo.head.commit.hexsha
    return path, commit, repo.is_dirty()


def load_sources_from_db(
    experiment: Dict,
    collection: 'Collection',
    to_directory: Union[str, Path],
    remove_src_directory: bool = SETTINGS.CODE_CHECKPOINT_REMOVE_SRC_DIRECTORY,
):
    import gridfs

    db = collection.database
    fs = gridfs.GridFS(db)
    if 'source_files' not in experiment['seml']:
        raise MongoDBError(
            f'No source files found for experiment with ID {experiment["_id"]}.'
        )
    source_files = experiment['seml']['source_files']
    target_directory = Path(to_directory)
    for path, _id in source_files:
        path = cast(str, path)
        # For the imports to prefer our loaded seml version, we need to convert the src-layout to the flat-layout.
        # https://packaging.python.org/en/latest/discussions/src-layout-vs-flat-layout/
        if remove_src_directory:
            path = src_layout_to_flat_layout(path)
        out_path = target_directory / path
        # only current user can read, write, or execute
        out_path.parent.mkdir(mode=0o700, parents=True, exist_ok=True)
        db_file = fs.find_one(_id)
        if db_file is None:
            raise MongoDBError(
                f"Could not find source file with ID '{_id}' for experiment with ID {experiment['_id']}."
            )
        with open(out_path, 'wb') as f:
            f.write(db_file.read())


def delete_batch_sources(collection: 'Collection', batch_id: int):
    db = collection.database
    filter_dict = {
        'metadata.batch_id': batch_id,
        'metadata.collection_name': f'{collection.name}',
    }
    source_files = db['fs.files'].find(filter_dict, {'_id'})
    source_files = [x['_id'] for x in source_files]
    if len(source_files) > 0:
        logging.info(
            f'Deleting {len(source_files)} source files corresponding '
            f'to batch {batch_id} in collection {collection.name}.'
        )
        delete_files(db, source_files)


def delete_orphaned_sources(collection: 'Collection', batch_ids=None):
    if batch_ids is not None:
        # check for empty batches within list of batch ids
        filter_dict = {'batch_id': {'$in': list(batch_ids)}}
    else:
        # check for any empty batches
        filter_dict = {}
    db_results = collection.find(filter_dict, {'batch_id'})
    remaining_batch_ids = {x['batch_id'] for x in db_results}
    empty_batch_ids = set(batch_ids) - remaining_batch_ids
    for b_id in empty_batch_ids:
        delete_batch_sources(collection, b_id)
